from notionary.file_upload.validation.port import FileUploadValidator
from notionary.utils.decorators import time_execution_async
from notionary.utils.mixins.logging import LoggingMixin


class FileUploadValidationService(LoggingMixin):
    def __init__(self) -> None:
        self._validators: list[FileUploadValidator] = []

    def register(self, validator: FileUploadValidator) -> None:
        self._validators.append(validator)

    @time_execution_async()
    async def validate(self) -> None:
        for validator in self._validators:
            self.logger.info("Validating with %s", validator.__class__.__name__)
            await validator.validate()



import re
from typing import override

from notionary.blocks.rich_text.markdown_rich_text_converter import (
    MarkdownRichTextConverter,
)
from notionary.blocks.rich_text.models import RichText
from notionary.blocks.schemas import CodeData, CodingLanguage, CreateCodeBlock
from notionary.page.content.parser.parsers.base import BlockParsingContext, LineParser
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class CodeParser(LineParser):
    DEFAULT_LANGUAGE = CodingLanguage.PLAIN_TEXT

    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry,
        rich_text_converter: MarkdownRichTextConverter,
    ) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_code_syntax()
        self._rich_text_converter = rich_text_converter
        self._code_start_pattern = self._syntax.regex_pattern
        self._code_end_pattern = self._syntax.end_regex_pattern or re.compile(
            r"^```\s*$"
        )

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False
        return self._is_code_fence_start(context.line)

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        code_lines = self._collect_code_lines(context)
        lines_consumed = self._count_lines_consumed(context)

        block = await self._create_code_block(
            opening_line=context.line, code_lines=code_lines
        )
        if not block:
            return

        context.lines_consumed = lines_consumed
        context.result_blocks.append(block)

    def _is_code_fence_start(self, line: str) -> bool:
        return self._code_start_pattern.match(line) is not None

    def _is_code_fence_end(self, line: str) -> bool:
        return self._code_end_pattern.match(line) is not None

    def _collect_code_lines(self, context: BlockParsingContext) -> list[str]:
        code_lines = []
        for line in context.get_remaining_lines():
            if self._is_code_fence_end(line):
                break
            code_lines.append(line)
        return code_lines

    def _count_lines_consumed(self, context: BlockParsingContext) -> int:
        for line_index, line in enumerate(context.get_remaining_lines()):
            if self._is_code_fence_end(line):
                return line_index + 1
        return len(context.get_remaining_lines())

    async def _create_code_block(
        self, opening_line: str, code_lines: list[str]
    ) -> CreateCodeBlock | None:
        match = self._code_start_pattern.match(opening_line)
        if not match:
            return None

        language = self._parse_language(match.group(1))
        rich_text = await self._create_rich_text_from_code(code_lines)

        code_data = CodeData(rich_text=rich_text, language=language, caption=[])
        return CreateCodeBlock(code=code_data)

    def _parse_language(self, language_str: str | None) -> CodingLanguage:
        return CodingLanguage.from_string(language_str, default=self.DEFAULT_LANGUAGE)

    async def _create_rich_text_from_code(
        self, code_lines: list[str]
    ) -> list[RichText]:
        content = "\n".join(code_lines) if code_lines else ""
        return await self._rich_text_converter.to_rich_text(content)

    def _is_code_fence_start(self, line: str) -> bool:
        return self._code_start_pattern.match(line) is not None

    def _is_code_fence_end(self, line: str) -> bool:
        return self._code_end_pattern.match(line) is not None



from enum import StrEnum


class RichTextPatterns(StrEnum):
    BOLD = r"\*\*(.+?)\*\*"
    """Matches **bold text**. Example: `**Hello**` → bolded "Hello"."""

    ITALIC = r"\*(.+?)\*"
    """Matches *italic text*. Example: `*Hello*` → italic "Hello"."""

    ITALIC_UNDERSCORE = r"_([^_]+?)_"
    """Matches _italic text_ using underscores. Example: `_Hello_` → italic "Hello"."""

    UNDERLINE = r"__(.+?)__"
    """Matches __underlined text__. Example: `__Hello__` → underlined "Hello"."""

    STRIKETHROUGH = r"~~(.+?)~~"
    """Matches ~~strikethrough~~ text. Example: `~~Hello~~` → struck-through "Hello"."""

    CODE = r"`(.+?)`"
    """Matches inline code. Example: `` `print("Hi")` `` → code span `print("Hi")`."""

    LINK = r"\[(.+?)\]\((.+?)\)"
    """Matches a hyperlink. Example: `[Google](https://google.com)` → link with label "Google"."""

    INLINE_EQUATION = r"\$(.+?)\$"
    """Matches an inline LaTeX equation. Example: `$E=mc^2$` → math formula E=mc²."""

    COLOR = r"\((\w+):(.+?)\)"
    """Matches colored text. Example: `(red:Important)` → text "Important" colored red."""

    PAGE_MENTION = r"@page\[([^\]]+)\]"
    """Matches a Notion page mention by name or ID. Example: `@page[My Page]`."""

    DATABASE_MENTION = r"@database\[([^\]]+)\]"
    """Matches a Notion database mention by name or ID. Example: `@database[Tasks]`."""

    DATASOURCE_MENTION = r"@datasource\[([^\]]+)\]"
    """Matches a Notion data source mention by name or ID. Example: `@datasource[My Data]`."""

    USER_MENTION = r"@user\[([^\]]+)\]"
    """Matches a Notion user mention by name or ID. Example: `@user[Some Person]`."""



from typing import Self

from notionary.workspace.query.models import (
    SortDirection,
    SortTimestamp,
    WorkspaceQueryConfig,
    WorkspaceQueryObjectType,
)


class NotionWorkspaceQueryConfigBuilder:
    def __init__(self, config: WorkspaceQueryConfig = None) -> None:
        self.config = config or WorkspaceQueryConfig()

    def with_query(self, query: str) -> Self:
        self.config.query = query
        return self

    def with_total_results_limit(self, limit: int) -> Self:
        self.config.total_results_limit = limit
        return self

    def with_pages_only(self) -> Self:
        self.config.object_type = WorkspaceQueryObjectType.PAGE
        return self

    def with_data_sources_only(self) -> Self:
        self.config.object_type = WorkspaceQueryObjectType.DATA_SOURCE
        return self

    def with_sort_direction(self, direction: SortDirection) -> Self:
        self.config.sort_direction = direction
        return self

    def with_sort_ascending(self) -> Self:
        return self.with_sort_direction(SortDirection.ASCENDING)

    def with_sort_descending(self) -> Self:
        return self.with_sort_direction(SortDirection.DESCENDING)

    def with_sort_timestamp(self, timestamp: SortTimestamp) -> Self:
        self.config.sort_timestamp = timestamp
        return self

    def with_sort_by_created_time(self) -> Self:
        return self.with_sort_timestamp(SortTimestamp.CREATED_TIME)

    def with_sort_by_last_edited(self) -> Self:
        return self.with_sort_timestamp(SortTimestamp.LAST_EDITED_TIME)

    def with_sort_by_created_time_ascending(self) -> Self:
        self.config.sort_timestamp = SortTimestamp.CREATED_TIME
        self.config.sort_direction = SortDirection.ASCENDING
        return self

    def with_sort_by_created_time_descending(self) -> Self:
        self.config.sort_timestamp = SortTimestamp.CREATED_TIME
        self.config.sort_direction = SortDirection.DESCENDING
        return self

    def with_sort_by_last_edited_ascending(self) -> Self:
        self.config.sort_timestamp = SortTimestamp.LAST_EDITED_TIME
        self.config.sort_direction = SortDirection.ASCENDING
        return self

    def with_sort_by_last_edited_descending(self) -> Self:
        self.config.sort_timestamp = SortTimestamp.LAST_EDITED_TIME
        self.config.sort_direction = SortDirection.DESCENDING
        return self

    def with_page_size(self, size: int) -> Self:
        self.config.page_size = min(size, 100)
        return self

    def with_start_cursor(self, cursor: str | None) -> Self:
        self.config.start_cursor = cursor
        return self

    def without_cursor(self) -> Self:
        self.config.start_cursor = None
        return self

    def build(self) -> WorkspaceQueryConfig:
        return self.config



from typing import TypeVar, override

from notionary.http.client import NotionHttpClient
from notionary.shared.entity.entity_metadata_update_client import (
    EntityMetadataUpdateClient,
)
from notionary.shared.entity.schemas import EntityResponseDto, NotionEntityUpdateDto

ResponseType = TypeVar("ResponseType", bound=EntityResponseDto)


class GenericEntityMetadataUpdateClient(NotionHttpClient, EntityMetadataUpdateClient):
    def __init__(
        self,
        entity_id: str,
        path_segment: str,
        response_dto_class: type[ResponseType],
        timeout: int = 30,
    ) -> None:
        super().__init__(timeout)
        self._entity_id = entity_id
        self._path_segment = path_segment
        self._response_dto_class = response_dto_class

    @override
    async def patch_metadata(self, updated_data: NotionEntityUpdateDto) -> ResponseType:
        updated_data_dict = updated_data.model_dump(
            exclude_unset=True, exclude_none=True
        )
        url = f"{self._path_segment}/{self._entity_id}"

        response_dict = await self.patch(url, data=updated_data_dict)
        return self._response_dto_class.model_validate(response_dict)



from dataclasses import dataclass

from notionary.shared.properties.type import PropertyType


@dataclass(frozen=True)
class PropertyTypeDescriptor:
    display_name: str
    description: str


class DatabasePropertyTypeDescriptorRegistry:
    def __init__(self):
        self._DESCRIPTORS = {
            PropertyType.TITLE: PropertyTypeDescriptor(
                display_name="Title",
                description="Required field for the main heading of the entry",
            ),
            PropertyType.RICH_TEXT: PropertyTypeDescriptor(
                display_name="Rich Text",
                description="Free-form text field for additional information",
            ),
            PropertyType.NUMBER: PropertyTypeDescriptor(
                display_name="Number", description="Numeric value field"
            ),
            PropertyType.CHECKBOX: PropertyTypeDescriptor(
                display_name="Checkbox", description="Boolean value (true/false)"
            ),
            PropertyType.DATE: PropertyTypeDescriptor(
                display_name="Date", description="Date or date range field"
            ),
            PropertyType.URL: PropertyTypeDescriptor(
                display_name="URL", description="Web address field"
            ),
            PropertyType.EMAIL: PropertyTypeDescriptor(
                display_name="Email", description="Email address field"
            ),
            PropertyType.PHONE_NUMBER: PropertyTypeDescriptor(
                display_name="Phone Number", description="Phone number field"
            ),
            PropertyType.FILES: PropertyTypeDescriptor(
                display_name="Files & Media", description="Upload or link to files"
            ),
            PropertyType.PEOPLE: PropertyTypeDescriptor(
                display_name="People", description="Reference to Notion users"
            ),
            PropertyType.SELECT: PropertyTypeDescriptor(
                display_name="Single Select",
                description="Choose one option from available choices",
            ),
            PropertyType.MULTI_SELECT: PropertyTypeDescriptor(
                display_name="Multi Select",
                description="Choose multiple options from available choices",
            ),
            PropertyType.STATUS: PropertyTypeDescriptor(
                display_name="Status",
                description="Track status with predefined options",
            ),
            PropertyType.RELATION: PropertyTypeDescriptor(
                display_name="Relation",
                description="Link to entries in another database",
            ),
            PropertyType.CREATED_TIME: PropertyTypeDescriptor(
                display_name="Created Time",
                description="Automatically set when the page is created",
            ),
            PropertyType.CREATED_BY: PropertyTypeDescriptor(
                display_name="Created By",
                description="Automatically set to the user who created the page",
            ),
            PropertyType.LAST_EDITED_TIME: PropertyTypeDescriptor(
                display_name="Last Edited Time",
                description="Automatically updated when the page is modified",
            ),
            PropertyType.LAST_EDITED_BY: PropertyTypeDescriptor(
                display_name="Last Edited By",
                description="Automatically set to the user who last edited the page",
            ),
            PropertyType.LAST_VISITED_TIME: PropertyTypeDescriptor(
                display_name="Last Visited Time",
                description="Automatically updated when the page is visited",
            ),
            PropertyType.FORMULA: PropertyTypeDescriptor(
                display_name="Formula",
                description="Computed value based on other properties",
            ),
            PropertyType.ROLLUP: PropertyTypeDescriptor(
                display_name="Rollup",
                description="Aggregate values from related database entries",
            ),
            PropertyType.BUTTON: PropertyTypeDescriptor(
                display_name="Button",
                description="Interactive button that triggers an action",
            ),
            PropertyType.LOCATION: PropertyTypeDescriptor(
                display_name="Location",
                description="Geographic location field",
            ),
            PropertyType.PLACE: PropertyTypeDescriptor(
                display_name="Place",
                description="Place or venue information",
            ),
            PropertyType.VERIFICATION: PropertyTypeDescriptor(
                display_name="Verification",
                description="Verification status field",
            ),
            PropertyType.UNIQUE_ID: PropertyTypeDescriptor(
                display_name="Unique ID",
                description="Auto-generated unique identifier",
            ),
        }

    def get_descriptor(self, property_type: PropertyType) -> PropertyTypeDescriptor:
        return self._DESCRIPTORS.get(
            property_type,
            PropertyTypeDescriptor(
                display_name=self._format_unknown_type_name(property_type),
                description="",
            ),
        )

    def _format_unknown_type_name(self, property_type: PropertyType) -> str:
        return property_type.value.replace("_", " ").title()



import random
from abc import ABC, abstractmethod
from collections.abc import Sequence
from pathlib import Path
from typing import Self, cast

from notionary.file_upload.service import NotionFileUpload
from notionary.shared.entity.entity_metadata_update_client import (
    EntityMetadataUpdateClient,
)
from notionary.shared.entity.schemas import EntityResponseDto
from notionary.shared.models.file import ExternalFile, FileType, NotionHostedFile
from notionary.shared.models.icon import EmojiIcon, IconType
from notionary.shared.models.parent import ParentType
from notionary.user.base import BaseUser
from notionary.user.service import UserService
from notionary.utils.mixins.logging import LoggingMixin
from notionary.utils.uuid_utils import extract_uuid


class Entity(LoggingMixin, ABC):
    def __init__(
        self,
        dto: EntityResponseDto,
        user_service: UserService | None = None,
        file_upload_service: NotionFileUpload | None = None,
    ) -> None:
        self._id = dto.id
        self._created_time = dto.created_time
        self._created_by = dto.created_by
        self._last_edited_time = dto.last_edited_time
        self._last_edited_by = dto.last_edited_by
        self._in_trash = dto.in_trash
        self._parent = dto.parent
        self._url = dto.url
        self._public_url = dto.public_url

        self._emoji_icon = self._extract_emoji_icon(dto)
        self._external_icon_url = self._extract_external_icon_url(dto)
        self._cover_image_url = self._extract_cover_image_url(dto)

        self._user_service = user_service or UserService()
        self._file_upload_service = file_upload_service or NotionFileUpload()

    @staticmethod
    def _extract_emoji_icon(dto: EntityResponseDto) -> str | None:
        if dto.icon is None:
            return None
        if dto.icon.type is not IconType.EMOJI:
            return None

        emoji_icon = cast(EmojiIcon, dto.icon)
        return emoji_icon.emoji

    @staticmethod
    def _extract_external_icon_url(dto: EntityResponseDto) -> str | None:
        if dto.icon is None:
            return None

        if dto.icon.type == IconType.EXTERNAL:
            external_icon = cast(ExternalFile, dto.icon)
            return external_icon.external.url
        elif dto.icon.type == IconType.FILE:
            notion_file_icon = cast(NotionHostedFile, dto.icon)
            return notion_file_icon.file.url

        return None

    @staticmethod
    def _extract_cover_image_url(dto: EntityResponseDto) -> str | None:
        if dto.cover is None:
            return None

        if dto.cover.type == FileType.EXTERNAL:
            external_cover = cast(ExternalFile, dto.cover)
            return external_cover.external.url
        elif dto.cover.type == FileType.FILE:
            notion_file_cover = cast(NotionHostedFile, dto.cover)
            return notion_file_cover.file.url

        return None

    @classmethod
    @abstractmethod
    async def from_id(cls, id: str) -> Self:
        pass

    @classmethod
    @abstractmethod
    async def from_title(cls, title: str) -> Self:
        pass

    @classmethod
    async def from_url(cls, url: str) -> Self:
        entity_id = extract_uuid(url)
        if not entity_id:
            raise ValueError(f"Could not extract entity ID from URL: {url}")
        return await cls.from_id(entity_id)

    @property
    @abstractmethod
    def _entity_metadata_update_client(self) -> EntityMetadataUpdateClient: ...

    @property
    def id(self) -> str:
        return self._id

    @property
    def created_time(self) -> str:
        return self._created_time

    @property
    def last_edited_time(self) -> str:
        return self._last_edited_time

    @property
    def in_trash(self) -> bool:
        return self._in_trash

    @property
    def emoji_icon(self) -> str | None:
        return self._emoji_icon

    @property
    def external_icon_url(self) -> str | None:
        return self._external_icon_url

    @property
    def cover_image_url(self) -> str | None:
        return self._cover_image_url

    @property
    def url(self) -> str:
        return self._url

    @property
    def public_url(self) -> str | None:
        return self._public_url

    def get_parent_database_id_if_present(self) -> str | None:
        if self._parent.type == ParentType.DATABASE_ID:
            return self._parent.database_id
        return None

    def get_parent_data_source_id_if_present(self) -> str | None:
        if self._parent.type == ParentType.DATA_SOURCE_ID:
            return self._parent.data_source_id
        return None

    def get_parent_page_id_if_present(self) -> str | None:
        if self._parent.type == ParentType.PAGE_ID:
            return self._parent.page_id
        return None

    def get_parent_block_id_if_present(self) -> str | None:
        if self._parent.type == ParentType.BLOCK_ID:
            return self._parent.block_id
        return None

    def is_workspace_parent(self) -> bool:
        return self._parent.type == ParentType.WORKSPACE

    async def get_created_by_user(self) -> BaseUser | None:
        return await self._user_service.get_user_by_id(self._created_by.id)

    async def get_last_edited_by_user(self) -> BaseUser | None:
        return await self._user_service.get_user_by_id(self._last_edited_by.id)

    async def set_emoji_icon(self, emoji: str) -> None:
        entity_response = await self._entity_metadata_update_client.patch_emoji_icon(
            emoji
        )
        self._emoji_icon = self._extract_emoji_icon(entity_response)
        self._external_icon_url = None

    async def set_external_icon(self, icon_url: str) -> None:
        entity_response = await self._entity_metadata_update_client.patch_external_icon(
            icon_url
        )
        self._emoji_icon = None
        self._external_icon_url = self._extract_external_icon_url(entity_response)

    async def set_icon_from_file(
        self, file_path: Path, filename: str | None = None
    ) -> None:
        upload_response = await self._file_upload_service.upload_file(
            file_path, filename
        )
        await self.set_icon_from_file_upload(upload_response.id)

    async def set_icon_from_bytes(
        self, file_content: bytes, filename: str, content_type: str | None = None
    ) -> None:
        upload_response = await self._file_upload_service.upload_from_bytes(
            file_content, filename, content_type
        )
        await self.set_icon_from_file_upload(upload_response.id)

    async def set_icon_from_file_upload(self, file_upload_id: str) -> None:
        entity_response = (
            await self._entity_metadata_update_client.patch_icon_from_file_upload(
                file_upload_id
            )
        )
        self._emoji_icon = None
        self._external_icon_url = self._extract_external_icon_url(entity_response)

    async def remove_icon(self) -> None:
        await self._entity_metadata_update_client.remove_icon()
        self._emoji_icon = None
        self._external_icon_url = None

    async def set_cover_image_by_url(self, image_url: str) -> None:
        entity_response = (
            await self._entity_metadata_update_client.patch_external_cover(image_url)
        )
        self._cover_image_url = self._extract_cover_image_url(entity_response)

    async def set_cover_image_from_file(
        self, file_path: Path, filename: str | None = None
    ) -> None:
        upload_response = await self._file_upload_service.upload_file(
            file_path, filename
        )
        await self.set_cover_image_from_file_upload(upload_response.id)

    async def set_cover_image_from_bytes(
        self, file_content: bytes, filename: str, content_type: str | None = None
    ) -> None:
        upload_response = await self._file_upload_service.upload_from_bytes(
            file_content, filename, content_type
        )
        await self.set_cover_image_from_file_upload(upload_response.id)

    async def set_cover_image_from_file_upload(self, file_upload_id: str) -> None:
        entity_response = (
            await self._entity_metadata_update_client.patch_cover_from_file_upload(
                file_upload_id
            )
        )
        self._cover_image_url = self._extract_cover_image_url(entity_response)

    async def set_random_gradient_cover(self) -> None:
        random_cover_url = self._get_random_gradient_cover()
        await self.set_cover_image_by_url(random_cover_url)

    async def remove_cover_image(self) -> None:
        await self._entity_metadata_update_client.remove_cover()
        self._cover_image_url = None

    async def move_to_trash(self) -> None:
        if self._in_trash:
            self.logger.warning("Entity is already in trash.")
            return

        entity_response = await self._entity_metadata_update_client.move_to_trash()
        self._in_trash = entity_response.in_trash

    async def restore_from_trash(self) -> None:
        if not self._in_trash:
            self.logger.warning("Entity is not in trash.")
            return

        entity_response = await self._entity_metadata_update_client.restore_from_trash()
        self._in_trash = entity_response.in_trash

    def __repr__(self) -> str:
        attrs = []
        for key, value in self.__dict__.items():
            if key.startswith("_") and not key.startswith("__"):
                attr_name = key[1:]
                attrs.append(f"{attr_name}={value!r}")

        attrs_str = ", ".join(attrs)
        return f"{self.__class__.__name__}({attrs_str})"

    def _get_random_gradient_cover(self) -> str:
        DEFAULT_NOTION_COVERS: Sequence[str] = [
            f"https://www.notion.so/images/page-cover/gradients_{i}.png"
            for i in range(1, 10)
        ]

        return random.choice(DEFAULT_NOTION_COVERS)



from typing import Any

from pydantic import BaseModel

from notionary.blocks.rich_text.models import RichText
from notionary.http.client import NotionHttpClient
from notionary.page.properties.schemas import (
    DateValue,
    PageCheckboxProperty,
    PageDateProperty,
    PageEmailProperty,
    PageMultiSelectProperty,
    PageNumberProperty,
    PagePhoneNumberProperty,
    PagePropertyT,
    PageRelationProperty,
    PageRichTextProperty,
    PageSelectProperty,
    PageStatusProperty,
    PageTitleProperty,
    PageURLProperty,
    RelationItem,
    SelectOption,
    StatusOption,
)
from notionary.page.schemas import NotionPageDto, PgePropertiesUpdateDto


class PagePropertyHttpClient(NotionHttpClient):
    def __init__(self, page_id: str) -> None:
        super().__init__()
        self._page_id = page_id

    async def patch_page(self, data: BaseModel) -> NotionPageDto:
        data_dict = data.model_dump(exclude_unset=True, exclude_none=True)

        result_dict = await self.patch(f"pages/{self._page_id}", data=data_dict)
        return NotionPageDto.model_validate(result_dict)

    async def _patch_property(
        self,
        property_name: str,
        value: Any,
        property_type: type[PagePropertyT],
        current_property: PagePropertyT | None = None,
    ) -> NotionPageDto:
        updated_property = self._create_updated_property(
            property_type, current_property, value
        )

        properties = {property_name: updated_property}
        update_dto = PgePropertiesUpdateDto(properties=properties)

        return await self.patch_page(update_dto)

    async def patch_title(self, property_name: str, title: str) -> NotionPageDto:
        return await self._patch_property(property_name, title, PageTitleProperty)

    async def patch_rich_text_property(
        self, property_name: str, text: str
    ) -> NotionPageDto:
        return await self._patch_property(property_name, text, PageRichTextProperty)

    async def patch_url_property(self, property_name: str, url: str) -> NotionPageDto:
        return await self._patch_property(property_name, url, PageURLProperty)

    async def patch_email_property(
        self, property_name: str, email: str
    ) -> NotionPageDto:
        return await self._patch_property(property_name, email, PageEmailProperty)

    async def patch_phone_property(
        self, property_name: str, phone: str
    ) -> NotionPageDto:
        return await self._patch_property(property_name, phone, PagePhoneNumberProperty)

    async def patch_number_property(
        self, property_name: str, number: int | float
    ) -> NotionPageDto:
        return await self._patch_property(property_name, number, PageNumberProperty)

    async def patch_checkbox_property(
        self, property_name: str, checked: bool
    ) -> NotionPageDto:
        return await self._patch_property(property_name, checked, PageCheckboxProperty)

    async def patch_select_property(
        self, property_name: str, value: str
    ) -> NotionPageDto:
        return await self._patch_property(property_name, value, PageSelectProperty)

    async def patch_multi_select_property(
        self, property_name: str, values: list[str]
    ) -> NotionPageDto:
        return await self._patch_property(
            property_name, values, PageMultiSelectProperty
        )

    async def patch_date_property(
        self, property_name: str, date_value: str | dict
    ) -> NotionPageDto:
        return await self._patch_property(property_name, date_value, PageDateProperty)

    async def patch_status_property(
        self, property_name: str, status: str
    ) -> NotionPageDto:
        return await self._patch_property(property_name, status, PageStatusProperty)

    async def patch_relation_property(
        self, property_name: str, relation_ids: str | list[str]
    ) -> NotionPageDto:
        if isinstance(relation_ids, str):
            relation_ids = [relation_ids]
        return await self._patch_property(
            property_name, relation_ids, PageRelationProperty
        )

    # TODO: Fix this shit here
    def _create_updated_property(
        self,
        property_type: type[PagePropertyT],
        current_property: PagePropertyT | None,
        value: Any,
    ) -> PagePropertyT:
        """
        Creates an updated property instance based on the property type and new value.
        """
        # Get the property ID from the current property if it exists
        property_id = current_property.id if current_property else ""

        if property_type == PageTitleProperty:
            return PageTitleProperty(
                id=property_id,
                title=[RichText(type="text", text={"content": str(value)})],
            )
        elif property_type == PageRichTextProperty:
            return PageRichTextProperty(
                id=property_id,
                rich_text=[RichText(type="text", text={"content": str(value)})],
            )
        elif property_type == PageURLProperty:
            return PageURLProperty(id=property_id, url=str(value))
        elif property_type == PageEmailProperty:
            return PageEmailProperty(id=property_id, email=str(value))
        elif property_type == PagePhoneNumberProperty:
            return PagePhoneNumberProperty(id=property_id, phone_number=str(value))
        elif property_type == PageNumberProperty:
            return PageNumberProperty(id=property_id, number=float(value))
        elif property_type == PageCheckboxProperty:
            return PageCheckboxProperty(id=property_id, checkbox=bool(value))
        elif property_type == PageSelectProperty:
            select_option = SelectOption(name=str(value))
            return PageSelectProperty(id=property_id, select=select_option)
        elif property_type == PageMultiSelectProperty:
            multi_select_options = [
                SelectOption(id="", name=str(item)) for item in value
            ]
            return PageMultiSelectProperty(
                id=property_id, multi_select=multi_select_options
            )
        elif property_type == PageDateProperty:
            if isinstance(value, dict) and "start" in value:
                date_value = DateValue(**value)
            else:
                date_value = DateValue(start=str(value))
            return PageDateProperty(id=property_id, date=date_value)
        elif property_type == PageStatusProperty:
            status_option = StatusOption(id="", name=str(value))
            return PageStatusProperty(id=property_id, status=status_option)
        elif property_type == PageRelationProperty:
            relation_items = [RelationItem(id=str(item)) for item in value]
            return PageRelationProperty(id=property_id, relation=relation_items)
        else:
            raise ValueError(f"Unsupported property type: {property_type}")



from enum import StrEnum
from typing import Protocol

from pydantic import BaseModel, Field, field_validator, model_serializer

from notionary.shared.typings import JsonDict


class SearchableEntity(Protocol):
    title: str


class SortDirection(StrEnum):
    ASCENDING = "ascending"
    DESCENDING = "descending"


class SortTimestamp(StrEnum):
    LAST_EDITED_TIME = "last_edited_time"
    CREATED_TIME = "created_time"


class WorkspaceQueryObjectType(StrEnum):
    PAGE = "page"
    DATA_SOURCE = "data_source"


class WorkspaceQueryConfig(BaseModel):
    query: str | None = None
    object_type: WorkspaceQueryObjectType | None = None
    sort_direction: SortDirection = SortDirection.DESCENDING
    sort_timestamp: SortTimestamp = SortTimestamp.LAST_EDITED_TIME

    page_size: int = Field(default=100, ge=1, le=100)
    start_cursor: str | None = None

    total_results_limit: int | None = None

    @field_validator("query")
    @classmethod
    def replace_empty_query_with_none(cls, value: str | None) -> str | None:
        if value is not None and not value.strip():
            return None
        return value

    @model_serializer
    def to_api_params(self) -> JsonDict:
        search_dict: JsonDict = {}

        if self.query:
            search_dict["query"] = self.query

        if self.object_type:
            search_dict["filter"] = {
                "property": "object",
                "value": self.object_type.value,
            }

        search_dict["sort"] = {
            "direction": self.sort_direction.value,
            "timestamp": self.sort_timestamp.value,
        }

        search_dict["page_size"] = self.page_size

        if self.start_cursor:
            search_dict["start_cursor"] = self.start_cursor

        return search_dict



from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.markdown.nodes.container import ContainerNode
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class ToggleMarkdownNode(ContainerNode):
    def __init__(
        self,
        title: str,
        children: list[MarkdownNode] | None = None,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
    ):
        super().__init__(syntax_registry=syntax_registry)
        self.title = title
        self.children = children or []

    @override
    def to_markdown(self) -> str:
        toggle_delimiter = self._get_toggle_delimiter()
        result = f"{toggle_delimiter} {self.title}"
        result += self.render_children()
        return result

    def _get_toggle_delimiter(self) -> str:
        return self._syntax_registry.get_toggle_syntax().start_delimiter



from __future__ import annotations

from enum import StrEnum
from typing import Self


class BlockColor(StrEnum):
    BLUE = "blue"
    BLUE_BACKGROUND = "blue_background"
    BROWN = "brown"
    BROWN_BACKGROUND = "brown_background"
    DEFAULT = "default"
    GRAY = "gray"
    GRAY_BACKGROUND = "gray_background"
    GREEN = "green"
    GREEN_BACKGROUND = "green_background"
    ORANGE = "orange"
    ORANGE_BACKGROUND = "orange_background"
    YELLOW = "yellow"
    YELLOW_BACKGROUND = "yellow_background"
    PINK = "pink"
    PINK_BACKGROUND = "pink_background"
    PURPLE = "purple"
    PURPLE_BACKGROUND = "purple_background"
    RED = "red"
    RED_BACKGROUND = "red_background"
    DEFAULT_BACKGROUND = "default_background"


class BlockType(StrEnum):
    AUDIO = "audio"
    BOOKMARK = "bookmark"
    BREADCRUMB = "breadcrumb"
    BULLETED_LIST_ITEM = "bulleted_list_item"
    CALLOUT = "callout"
    CHILD_DATABASE = "child_database"
    CHILD_PAGE = "child_page"
    CODE = "code"
    COLUMN = "column"
    COLUMN_LIST = "column_list"
    DIVIDER = "divider"
    EMBED = "embed"
    EQUATION = "equation"
    FILE = "file"
    HEADING_1 = "heading_1"
    HEADING_2 = "heading_2"
    HEADING_3 = "heading_3"
    IMAGE = "image"
    LINK_PREVIEW = "link_preview"
    LINK_TO_PAGE = "link_to_page"
    NUMBERED_LIST_ITEM = "numbered_list_item"
    PARAGRAPH = "paragraph"
    PDF = "pdf"
    QUOTE = "quote"
    SYNCED_BLOCK = "synced_block"
    TABLE = "table"
    TABLE_OF_CONTENTS = "table_of_contents"
    TABLE_ROW = "table_row"
    TO_DO = "to_do"
    TOGGLE = "toggle"
    UNSUPPORTED = "unsupported"
    VIDEO = "video"


class CodingLanguage(StrEnum):
    ABAP = "abap"
    ARDUINO = "arduino"
    BASH = "bash"
    BASIC = "basic"
    C = "c"
    CLOJURE = "clojure"
    COFFEESCRIPT = "coffeescript"
    CPP = "c++"
    CSHARP = "c#"
    CSS = "css"
    DART = "dart"
    DIFF = "diff"
    DOCKER = "docker"
    ELIXIR = "elixir"
    ELM = "elm"
    ERLANG = "erlang"
    FLOW = "flow"
    FORTRAN = "fortran"
    FSHARP = "f#"
    GHERKIN = "gherkin"
    GLSL = "glsl"
    GO = "go"
    GRAPHQL = "graphql"
    GROOVY = "groovy"
    HASKELL = "haskell"
    HTML = "html"
    JAVA = "java"
    JAVASCRIPT = "javascript"
    JSON = "json"
    JULIA = "julia"
    KOTLIN = "kotlin"
    LATEX = "latex"
    LESS = "less"
    LISP = "lisp"
    LIVESCRIPT = "livescript"
    LUA = "lua"
    MAKEFILE = "makefile"
    MARKDOWN = "markdown"
    MARKUP = "markup"
    MATLAB = "matlab"
    MERMAID = "mermaid"
    NIX = "nix"
    OBJECTIVE_C = "objective-c"
    OCAML = "ocaml"
    PASCAL = "pascal"
    PERL = "perl"
    PHP = "php"
    PLAIN_TEXT = "plain text"
    POWERSHELL = "powershell"
    PROLOG = "prolog"
    PROTOBUF = "protobuf"
    PYTHON = "python"
    R = "r"
    REASON = "reason"
    RUBY = "ruby"
    RUST = "rust"
    SASS = "sass"
    SCALA = "scala"
    SCHEME = "scheme"
    SCSS = "scss"
    SHELL = "shell"
    SQL = "sql"
    SWIFT = "swift"
    TYPESCRIPT = "typescript"
    VB_NET = "vb.net"
    VERILOG = "verilog"
    VHDL = "vhdl"
    VISUAL_BASIC = "visual basic"
    WEBASSEMBLY = "webassembly"
    XML = "xml"
    YAML = "yaml"
    JAVA_C_CPP_CSHARP = "java/c/c++/c#"

    @classmethod
    def from_string(cls, lang_str: str | None, default: Self | None = None) -> Self:
        if not lang_str:
            return default if default is not None else cls.PLAIN_TEXT

        normalized = lang_str.lower().strip()

        aliases = {
            "cpp": cls.CPP,
            "c++": cls.CPP,
            "js": cls.JAVASCRIPT,
            "py": cls.PYTHON,
            "ts": cls.TYPESCRIPT,
        }

        if normalized in aliases:
            return aliases[normalized]

        for member in cls:
            if member.value.lower() == normalized:
                return member

        return default if default is not None else cls.PLAIN_TEXT


class VideoFileType(StrEnum):
    AMV = ".amv"
    ASF = ".asf"
    AVI = ".avi"
    F4V = ".f4v"
    FLV = ".flv"
    GIFV = ".gifv"
    MKV = ".mkv"
    MOV = ".mov"
    MPG = ".mpg"
    MPEG = ".mpeg"
    MPV = ".mpv"
    MP4 = ".mp4"
    M4V = ".m4v"
    QT = ".qt"
    WMV = ".wmv"

    @classmethod
    def get_all_extensions(cls) -> set[str]:
        return {ext.value for ext in cls}

    @classmethod
    def is_valid_extension(cls, filename: str) -> bool:
        lower_filename = filename.lower()
        return any(lower_filename.endswith(ext.value) for ext in cls)



from dataclasses import dataclass


@dataclass(frozen=True)
class MarkdownGrammar:
    spaces_per_nesting_level = 4
    numbered_list_placeholder = "__NUM__"
    column_delimiter = ":::"
    toggle_delimiter = "+++"
    table_delimiter = "|"



from typing import ClassVar

from notionary.data_source.properties.schemas import DataSourceProperty
from notionary.data_source.query.schema import (
    ArrayOperator,
    BooleanOperator,
    DateOperator,
    NumberOperator,
    Operator,
    SelectOperator,
    StringOperator,
)
from notionary.exceptions.data_source.builder import InvalidOperatorForPropertyType
from notionary.shared.properties.type import PropertyType


class QueryValidator:
    _PROPERTY_TYPE_OPERATORS: ClassVar[dict[PropertyType, list[type[Operator]]]] = {
        PropertyType.TITLE: [StringOperator],
        PropertyType.RICH_TEXT: [StringOperator],
        PropertyType.URL: [StringOperator],
        PropertyType.EMAIL: [StringOperator],
        PropertyType.PHONE_NUMBER: [StringOperator],
        PropertyType.SELECT: [SelectOperator],
        PropertyType.STATUS: [SelectOperator],
        PropertyType.MULTI_SELECT: [ArrayOperator],
        PropertyType.NUMBER: [NumberOperator],
        PropertyType.DATE: [DateOperator],
        PropertyType.CREATED_TIME: [DateOperator],
        PropertyType.LAST_EDITED_TIME: [DateOperator],
        PropertyType.PEOPLE: [ArrayOperator],
        PropertyType.CREATED_BY: [ArrayOperator],
        PropertyType.LAST_EDITED_BY: [ArrayOperator],
        PropertyType.RELATION: [ArrayOperator],
        PropertyType.CHECKBOX: [BooleanOperator],
    }

    def validate_operator_for_property(
        self, property_name: str, property_obj: DataSourceProperty, operator: Operator
    ) -> None:
        if not self._is_operator_valid_for_property_type(property_obj.type, operator):
            valid_operators = self._get_valid_operators_for_property_type(
                property_obj.type
            )
            raise InvalidOperatorForPropertyType(
                property_name=property_name,
                property_type=property_obj.type,
                operator=operator,
                valid_operators=valid_operators,
            )

    def _is_operator_valid_for_property_type(
        self, property_type: PropertyType, operator: Operator
    ) -> bool:
        allowed_operator_types = self._PROPERTY_TYPE_OPERATORS.get(property_type, [])
        valid_operator_values = self._get_operator_values_from_types(
            allowed_operator_types
        )
        return operator.value in valid_operator_values

    def _get_operator_values_from_types(
        self, operator_types: list[type[Operator]]
    ) -> set[str]:
        values: set[str] = set()
        for operator_type in operator_types:
            for operator in operator_type:
                values.add(operator.value)
        return values

    def _get_valid_operators_for_property_type(
        self, property_type: PropertyType
    ) -> list[Operator]:
        allowed_operator_types = self._PROPERTY_TYPE_OPERATORS.get(property_type, [])
        return self._collect_all_operators_from_types(allowed_operator_types)

    def _collect_all_operators_from_types(
        self, operator_types: list[type[Operator]]
    ) -> list[Operator]:
        operators: list[Operator] = []
        for operator_type in operator_types:
            operators.extend(self._get_all_enum_values(operator_type))
        return operators

    def _get_all_enum_values(self, operator_type: type[Operator]) -> list[Operator]:
        return list(operator_type)



from typing import override

from notionary.blocks.schemas import Block, BlockType
from notionary.page.content.renderer.renderers.captioned_block import (
    CaptionedBlockRenderer,
)


class EmbedRenderer(CaptionedBlockRenderer):
    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.EMBED

    @override
    async def _render_main_content(self, block: Block) -> str:
        url = self._extract_embed_url(block)

        if not url:
            return ""

        syntax = self._syntax_registry.get_embed_syntax()
        return f"{syntax.start_delimiter}{url}{syntax.end_delimiter}"

    def _extract_embed_url(self, block: Block) -> str:
        if not block.embed:
            return ""
        return block.embed.url or ""



from abc import ABC, abstractmethod


class NameIdResolver(ABC):
    @abstractmethod
    def resolve_name_to_id(self, name: str) -> str | None:
        pass

    @abstractmethod
    def resolve_id_to_name(self, id: str) -> str | None:
        pass



import logging
import os
from typing import ClassVar

from dotenv import load_dotenv

load_dotenv(override=True)

logger = logging.getLogger("notionary")
logger.addHandler(logging.NullHandler())


def configure_library_logging(level: str = "WARNING") -> None:
    log_level = getattr(logging, level.upper(), logging.WARNING)

    library_logger = logging.getLogger("notionary")

    if library_logger.handlers:
        library_logger.handlers.clear()

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )

    library_logger.setLevel(log_level)
    library_logger.addHandler(handler)

    _suppress_noisy_third_party_loggers()


def _suppress_noisy_third_party_loggers() -> None:
    noisy_loggers = ["httpx", "httpcore", "httpcore.connection", "httpcore.http11"]

    for logger_name in noisy_loggers:
        logging.getLogger(logger_name).setLevel(logging.WARNING)


def _auto_configure_from_environment() -> None:
    env_log_level = os.getenv("NOTIONARY_LOG_LEVEL")

    if env_log_level:
        configure_library_logging(env_log_level)


_auto_configure_from_environment()


class LoggingMixin:
    logger: ClassVar[logging.Logger] = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.logger = logging.getLogger(f"notionary.{cls.__name__}")

    @property
    def instance_logger(self) -> logging.Logger:
        if not hasattr(self, "_logger"):
            self._logger = logging.getLogger(f"notionary.{self.__class__.__name__}")
        return self._logger



from enum import StrEnum
from typing import Annotated, Literal

from pydantic import BaseModel, Field


class UserType(StrEnum):
    PERSON = "person"
    BOT = "bot"


class WorkspaceOwnerType(StrEnum):
    USER = "user"
    WORKSPACE = "workspace"


class PersonUserDto(BaseModel):
    email: str | None = None


class BotOwnerDto(BaseModel):
    type: WorkspaceOwnerType
    workspace: bool | None = None


class WorkspaceLimits(BaseModel):
    max_file_upload_size_in_bytes: int


class BotUserDto(BaseModel):
    owner: BotOwnerDto | None = None
    workspace_name: str | None = None
    workspace_limits: WorkspaceLimits | None = None


class NotionUserBase(BaseModel):
    object: Literal["user"] = "user"
    id: str

    type: UserType

    name: str | None = None
    avatar_url: str | None = None


class PersonUserResponseDto(NotionUserBase):
    type: Literal[UserType.PERSON] = UserType.PERSON
    person: PersonUserDto


class BotUserResponseDto(NotionUserBase):
    type: Literal[UserType.BOT] = UserType.BOT
    bot: BotUserDto


UserResponseDto = Annotated[
    PersonUserResponseDto | BotUserResponseDto, Field(discriminator="type")
]


class NotionUsersListResponse(BaseModel):
    results: list[UserResponseDto]
    next_cursor: str | None = None
    has_more: bool


class PartialUserDto(BaseModel):
    object: Literal["user"] = "user"
    id: str



from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

from notionary.exceptions.search import (
    DatabaseNotFound,
    DataSourceNotFound,
    PageNotFound,
)
from notionary.utils.fuzzy import find_all_matches
from notionary.workspace.client import WorkspaceClient
from notionary.workspace.query.builder import NotionWorkspaceQueryConfigBuilder
from notionary.workspace.query.models import SearchableEntity, WorkspaceQueryConfig

if TYPE_CHECKING:
    from notionary import NotionDatabase, NotionDataSource, NotionPage


class WorkspaceQueryService:
    def __init__(self, client: WorkspaceClient | None = None) -> None:
        self._client = client or WorkspaceClient()

    async def get_pages_stream(
        self, search_config: WorkspaceQueryConfig
    ) -> AsyncIterator[NotionPage]:
        from notionary import NotionPage

        async for page_dto in self._client.query_pages_stream(search_config):
            yield await NotionPage.from_id(page_dto.id)

    async def get_pages(self, search_config: WorkspaceQueryConfig) -> list[NotionPage]:
        from notionary import NotionPage

        page_dtos = [
            dto async for dto in self._client.query_pages_stream(search_config)
        ]
        page_tasks = [NotionPage.from_id(dto.id) for dto in page_dtos]
        return await asyncio.gather(*page_tasks)

    async def get_data_sources_stream(
        self, search_config: WorkspaceQueryConfig
    ) -> AsyncIterator[NotionDataSource]:
        from notionary import NotionDataSource

        async for data_source_dto in self._client.query_data_sources_stream(
            search_config
        ):
            yield await NotionDataSource.from_id(data_source_dto.id)

    async def get_data_sources(
        self, search_config: WorkspaceQueryConfig
    ) -> list[NotionDataSource]:
        from notionary import NotionDataSource

        data_source_dtos = [
            dto async for dto in self._client.query_data_sources_stream(search_config)
        ]
        data_source_tasks = [
            NotionDataSource.from_id(dto.id) for dto in data_source_dtos
        ]
        return await asyncio.gather(*data_source_tasks)

    async def find_data_source(self, query: str) -> NotionDataSource:
        config = (
            NotionWorkspaceQueryConfigBuilder()
            .with_query(query)
            .with_data_sources_only()
            .with_page_size(100)
            .build()
        )
        data_sources = await self.get_data_sources(config)
        return self._find_exact_match(data_sources, query, DataSourceNotFound)

    async def find_page(self, query: str) -> NotionPage:
        config = (
            NotionWorkspaceQueryConfigBuilder()
            .with_query(query)
            .with_pages_only()
            .with_page_size(100)
            .build()
        )
        pages = await self.get_pages(config)
        return self._find_exact_match(pages, query, PageNotFound)

    async def find_database(self, query: str) -> NotionDatabase:
        config = (
            NotionWorkspaceQueryConfigBuilder()
            .with_query(query)
            .with_data_sources_only()
            .with_page_size(100)
            .build()
        )
        data_sources = await self.get_data_sources(config)

        parent_database_ids = [
            data_sources.get_parent_database_id_if_present()
            for data_sources in data_sources
        ]
        # filter none values which should not happen but for safety
        parent_database_ids = [id for id in parent_database_ids if id is not None]

        parent_database_tasks = [
            NotionDatabase.from_id(db_id) for db_id in parent_database_ids
        ]
        parent_databases = await asyncio.gather(*parent_database_tasks)
        potential_databases = [
            database for database in parent_databases if database is not None
        ]

        return self._find_exact_match(potential_databases, query, DatabaseNotFound)

    def _find_exact_match(
        self,
        search_results: list[SearchableEntity],
        query: str,
        exception_class: type[Exception],
    ) -> SearchableEntity:
        if not search_results:
            raise exception_class(query, [])

        query_lower = query.lower()
        exact_matches = [
            result for result in search_results if result.title.lower() == query_lower
        ]

        if exact_matches:
            return exact_matches[0]

        suggestions = self._get_fuzzy_suggestions(search_results, query)
        raise exception_class(query, suggestions)

    def _get_fuzzy_suggestions(
        self, search_results: list[SearchableEntity], query: str
    ) -> list[str]:
        sorted_by_similarity = find_all_matches(
            query=query,
            items=search_results,
            text_extractor=lambda entity: entity.title,
            min_similarity=0.6,
        )

        if sorted_by_similarity:
            return [result.title for result in sorted_by_similarity[:5]]

        return [result.title for result in search_results[:5]]



from typing import override

from notionary.exceptions.file_upload import FilenameTooLongError
from notionary.file_upload.config.constants import NOTION_MAX_FILENAME_BYTES
from notionary.file_upload.validation.port import FileUploadValidator


class FileNameLengthValidator(FileUploadValidator):
    def __init__(self, filename: str) -> None:
        self._filename = filename
        self._max_filename_bytes = NOTION_MAX_FILENAME_BYTES

    @override
    async def validate(self) -> None:
        filename_bytes = len(self._filename.encode("utf-8"))
        if filename_bytes > self._max_filename_bytes:
            raise FilenameTooLongError(
                filename=self._filename,
                filename_bytes=filename_bytes,
                max_filename_bytes=self._max_filename_bytes,
            )



from notionary.blocks.schemas import Block, BlockChildrenResponse, BlockCreatePayload
from notionary.http.client import NotionHttpClient
from notionary.shared.typings import JsonDict
from notionary.utils.decorators import time_execution_async
from notionary.utils.pagination import paginate_notion_api


class NotionBlockHttpClient(NotionHttpClient):
    BATCH_SIZE = 100

    async def get_block(self, block_id: str) -> Block:
        response = await self.get(f"blocks/{block_id}")
        return Block.model_validate(response)

    async def delete_block(self, block_id: str) -> None:
        self.logger.debug("Deleting block: %s", block_id)
        await self.delete(f"blocks/{block_id}")

    @time_execution_async()
    async def get_block_tree(self, parent_block_id: str) -> list[Block]:
        blocks_at_this_level = await self.get_all_block_children(parent_block_id)

        for block in blocks_at_this_level:
            if block.has_children:
                nested_children = await self.get_block_tree(parent_block_id=block.id)
                block.children = nested_children

        return blocks_at_this_level

    @time_execution_async()
    async def get_all_block_children(self, parent_block_id: str) -> list[Block]:
        self.logger.debug("Retrieving all children for block: %s", parent_block_id)

        all_blocks = await paginate_notion_api(
            self.get_block_children, block_id=parent_block_id
        )

        self.logger.debug(
            "Retrieved %d total children for block %s", len(all_blocks), parent_block_id
        )
        return all_blocks

    async def get_block_children(
        self, block_id: str, start_cursor: str | None = None, page_size: int = 100
    ) -> BlockChildrenResponse:
        self.logger.debug("Retrieving children of block: %s", block_id)

        params = {"page_size": min(page_size, 100)}
        if start_cursor:
            params["start_cursor"] = start_cursor

        response = await self.get(f"blocks/{block_id}/children", params=params)
        return BlockChildrenResponse.model_validate(response)

    async def append_block_children(
        self,
        block_id: str,
        children: list[BlockCreatePayload],
        insert_after_block_id: str | None = None,
    ) -> BlockChildrenResponse | None:
        if not children:
            self.logger.warning("No children provided to append")
            return None

        self.logger.debug("Appending %d children to block: %s", len(children), block_id)

        batches = self._split_into_batches(children)

        if len(batches) == 1:
            children_dicts = self._serialize_blocks(batches[0])
            return await self._send_append_request(
                block_id, children_dicts, insert_after_block_id
            )

        return await self._send_batched_append_requests(
            block_id, batches, insert_after_block_id
        )

    def _split_into_batches(
        self, blocks: list[BlockCreatePayload]
    ) -> list[list[BlockCreatePayload]]:
        batches = []
        for i in range(0, len(blocks), self.BATCH_SIZE):
            batch = blocks[i : i + self.BATCH_SIZE]
            batches.append(batch)
        return batches

    def _serialize_blocks(self, blocks: list[BlockCreatePayload]) -> list[JsonDict]:
        return [block.model_dump(exclude_none=True) for block in blocks]

    async def _send_append_request(
        self, block_id: str, children: list[JsonDict], after_block_id: str | None = None
    ) -> BlockChildrenResponse:
        payload = {"children": children}
        if after_block_id:
            payload["after"] = after_block_id

        response = await self.patch(f"blocks/{block_id}/children", payload)
        return BlockChildrenResponse.model_validate(response)

    async def _send_batched_append_requests(
        self,
        block_id: str,
        batches: list[list[BlockCreatePayload]],
        initial_after_block_id: str | None = None,
    ) -> BlockChildrenResponse:
        total_blocks = sum(len(batch) for batch in batches)
        self.logger.info(
            "Appending %d blocks in %d batches", total_blocks, len(batches)
        )

        all_responses = []
        after_block_id = initial_after_block_id

        for batch_index, batch in enumerate(batches, start=1):
            self.logger.debug(
                "Processing batch %d/%d (%d blocks)",
                batch_index,
                len(batches),
                len(batch),
            )

            children_dicts = self._serialize_blocks(batch)
            response = await self._send_append_request(
                block_id, children_dicts, after_block_id
            )
            all_responses.append(response)

            if response.results:
                after_block_id = response.results[-1].id

            self.logger.debug("Completed batch %d/%d", batch_index, len(batches))

        self.logger.info("Successfully appended all blocks in %d batches", len(batches))
        return self._merge_responses(all_responses)

    def _merge_responses(
        self, responses: list[BlockChildrenResponse]
    ) -> BlockChildrenResponse:
        if not responses:
            raise ValueError(
                "Cannot merge empty response list - this should never happen"
            )

        first_response = responses[0]
        all_results = [block for response in responses for block in response.results]

        return BlockChildrenResponse(
            object=first_response.object,
            results=all_results,
            next_cursor=None,
            has_more=False,
            type=first_response.type,
            block=first_response.block,
            request_id=responses[-1].request_id,
        )



from pathlib import Path
from typing import ClassVar, override

from notionary.exceptions.file_upload import (
    NoFileExtensionException,
    UnsupportedFileTypeException,
)
from notionary.file_upload.validation.models import (
    AudioExtension,
    AudioMimeType,
    DocumentExtension,
    DocumentMimeType,
    FileCategory,
    ImageExtension,
    ImageMimeType,
    VideoExtension,
    VideoMimeType,
)
from notionary.file_upload.validation.port import FileUploadValidator


class FileExtensionValidator(FileUploadValidator):
    EXTENSION_TO_MIME: ClassVar[dict[str, str]] = {
        AudioExtension.AAC: AudioMimeType.AAC,
        AudioExtension.ADTS: AudioMimeType.AAC,
        AudioExtension.MID: AudioMimeType.MIDI,
        AudioExtension.MIDI: AudioMimeType.MIDI,
        AudioExtension.MP3: AudioMimeType.MPEG,
        AudioExtension.MPGA: AudioMimeType.MPEG,
        AudioExtension.M4A: AudioMimeType.MP4,
        AudioExtension.M4B: AudioMimeType.MP4,
        AudioExtension.MP4: AudioMimeType.MP4,
        AudioExtension.OGA: AudioMimeType.OGG,
        AudioExtension.OGG: AudioMimeType.OGG,
        AudioExtension.WAV: AudioMimeType.WAV,
        AudioExtension.WMA: AudioMimeType.WMA,
        DocumentExtension.PDF: DocumentMimeType.PDF,
        DocumentExtension.TXT: DocumentMimeType.PLAIN_TEXT,
        DocumentExtension.JSON: DocumentMimeType.JSON,
        DocumentExtension.DOC: DocumentMimeType.MSWORD,
        DocumentExtension.DOT: DocumentMimeType.MSWORD,
        DocumentExtension.DOCX: DocumentMimeType.WORD_DOCUMENT,
        DocumentExtension.DOTX: DocumentMimeType.WORD_TEMPLATE,
        DocumentExtension.XLS: DocumentMimeType.EXCEL,
        DocumentExtension.XLT: DocumentMimeType.EXCEL,
        DocumentExtension.XLA: DocumentMimeType.EXCEL,
        DocumentExtension.XLSX: DocumentMimeType.EXCEL_SHEET,
        DocumentExtension.XLTX: DocumentMimeType.EXCEL_TEMPLATE,
        DocumentExtension.PPT: DocumentMimeType.POWERPOINT,
        DocumentExtension.POT: DocumentMimeType.POWERPOINT,
        DocumentExtension.PPS: DocumentMimeType.POWERPOINT,
        DocumentExtension.PPA: DocumentMimeType.POWERPOINT,
        DocumentExtension.PPTX: DocumentMimeType.POWERPOINT_PRESENTATION,
        DocumentExtension.POTX: DocumentMimeType.POWERPOINT_TEMPLATE,
        ImageExtension.GIF: ImageMimeType.GIF,
        ImageExtension.HEIC: ImageMimeType.HEIC,
        ImageExtension.JPEG: ImageMimeType.JPEG,
        ImageExtension.JPG: ImageMimeType.JPEG,
        ImageExtension.PNG: ImageMimeType.PNG,
        ImageExtension.SVG: ImageMimeType.SVG,
        ImageExtension.TIF: ImageMimeType.TIFF,
        ImageExtension.TIFF: ImageMimeType.TIFF,
        ImageExtension.WEBP: ImageMimeType.WEBP,
        ImageExtension.ICO: ImageMimeType.ICON,
        VideoExtension.AMV: VideoMimeType.AMV,
        VideoExtension.ASF: VideoMimeType.ASF,
        VideoExtension.WMV: VideoMimeType.ASF,
        VideoExtension.AVI: VideoMimeType.AVI,
        VideoExtension.F4V: VideoMimeType.F4V,
        VideoExtension.FLV: VideoMimeType.FLV,
        VideoExtension.GIFV: VideoMimeType.WEBM,
        VideoExtension.M4V: VideoMimeType.MP4,
        VideoExtension.MP4: VideoMimeType.MP4,
        VideoExtension.MKV: VideoMimeType.MKV,
        VideoExtension.WEBM: VideoMimeType.WEBM,
        VideoExtension.MOV: VideoMimeType.QUICKTIME,
        VideoExtension.QT: VideoMimeType.QUICKTIME,
        VideoExtension.MPEG: VideoMimeType.MPEG,
    }

    EXTENSION_TO_CATEGORY: ClassVar[dict[str, FileCategory]] = {
        **{ext.value: FileCategory.AUDIO for ext in AudioExtension},
        **{ext.value: FileCategory.DOCUMENT for ext in DocumentExtension},
        **{ext.value: FileCategory.IMAGE for ext in ImageExtension},
        **{ext.value: FileCategory.VIDEO for ext in VideoExtension},
    }

    def __init__(self, filename: str | Path) -> None:
        self.filename = Path(filename).name if isinstance(filename, Path) else filename

    @override
    async def validate(self) -> None:
        extension = self._extract_extension(self.filename)

        if not extension:
            raise NoFileExtensionException(self.filename)

        if not self._is_supported(extension):
            supported_by_category = self._get_supported_extensions_by_category()
            raise UnsupportedFileTypeException(
                extension, self.filename, supported_by_category
            )

    @staticmethod
    def _extract_extension(filename: str) -> str:
        import os

        _, ext = os.path.splitext(filename)
        return ext.lower()

    def _is_supported(self, extension: str) -> bool:
        normalized = self._normalize_extension(extension)
        return normalized in self.EXTENSION_TO_MIME

    @staticmethod
    def _normalize_extension(extension: str) -> str:
        extension = extension.lower()
        if not extension.startswith("."):
            extension = f".{extension}"
        return extension

    def _get_supported_extensions_by_category(self) -> dict[str, list[str]]:
        result = {}
        for category in FileCategory:
            extensions = [
                ext
                for ext, cat in self.EXTENSION_TO_CATEGORY.items()
                if cat == category
            ]
            result[category.value] = extensions
        return result



from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class CaptionMarkdownNodeMixin:
    _syntax_registry: SyntaxDefinitionRegistry

    def _append_caption_to_markdown(
        self, base_markdown: str, caption: str | None
    ) -> str:
        if not caption:
            return base_markdown

        caption_syntax = self._syntax_registry.get_caption_syntax()
        return f"{base_markdown}\n{caption_syntax.start_delimiter} {caption}"



from notionary.exceptions.base import NotionaryException

RATIO_TOLERANCE = 0.0001


class InsufficientColumnsError(NotionaryException):
    def __init__(self, column_count: int) -> None:
        self.column_count = column_count
        super().__init__(
            f"Columns container must contain at least 2 column blocks, but only {column_count} found"
        )


class InvalidColumnRatioSumError(NotionaryException):
    def __init__(self, total: float, tolerance: float = RATIO_TOLERANCE) -> None:
        self.total = total
        self.tolerance = tolerance
        super().__init__(
            f"Width ratios must sum to 1.0 (±{tolerance}), but sum is {total}"
        )


class UnsupportedVideoFormatError(ValueError):
    def __init__(self, url: str, supported_formats: list[str]) -> None:
        self.url = url
        self.supported_formats = supported_formats
        super().__init__(self._create_user_friendly_message())

    def _create_user_friendly_message(self) -> str:
        formats = ", ".join(self.supported_formats[:5])
        remaining = len(self.supported_formats) - 5

        if remaining > 0:
            formats += f" and {remaining} more"

        return (
            f"The video URL '{self.url}' uses an unsupported format.\n"
            f"Supported formats include: {formats}.\n"
            f"YouTube embed and watch URLs are also supported."
            f"Also see https://developers.notion.com/reference/block#video for more information."
        )



from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.shared.entity.schemas import Describable, Titled


async def extract_title(
    entity: Titled,
    rich_text_converter: RichTextToMarkdownConverter,
) -> str:
    return await rich_text_converter.to_markdown(entity.title)


async def extract_description(
    entity: Describable,
    rich_text_converter: RichTextToMarkdownConverter,
) -> str | None:
    if not entity.description:
        return None
    return await rich_text_converter.to_markdown(entity.description)



from enum import StrEnum
from typing import Annotated, Literal

from pydantic import BaseModel, Field

from notionary.shared.models.file import (
    File,
)


class IconType(StrEnum):
    EMOJI = "emoji"
    EXTERNAL = "external"
    FILE = "file"
    FILE_UPLOAD = "file_upload"


class EmojiIcon(BaseModel):
    type: Literal[IconType.EMOJI] = IconType.EMOJI
    emoji: str


type Icon = Annotated[EmojiIcon | File, Field(discriminator="type")]


from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.markdown.nodes.container import ContainerNode
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class CalloutMarkdownNode(ContainerNode):
    def __init__(
        self,
        text: str,
        emoji: str | None = None,
        children: list[MarkdownNode] | None = None,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
    ):
        super().__init__(syntax_registry=syntax_registry)
        self.text = text
        self.emoji = emoji
        self.children = children or []

    @override
    def to_markdown(self) -> str:
        callout_content = self._format_callout_content()
        start_delimiter = self._syntax_registry.get_callout_syntax().start_delimiter
        result = f"{start_delimiter}({callout_content})"

        result += self.render_children()

        return result

    def _format_callout_content(self) -> str:
        if self.emoji:
            return f'{self.text} "{self.emoji}"'
        return self.text



from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class ParagraphMarkdownNode(MarkdownNode):
    def __init__(
        self, text: str, syntax_registry: SyntaxDefinitionRegistry | None = None
    ):
        super().__init__(syntax_registry=syntax_registry)
        self.text = text

    @override
    def to_markdown(self) -> str:
        return self.text



from abc import abstractmethod
from typing import override

from notionary.blocks.schemas import (
    Block,
    ExternalFileWithCaption,
    NotionHostedFileWithCaption,
)
from notionary.page.content.renderer.renderers.captioned_block import (
    CaptionedBlockRenderer,
)
from notionary.page.content.syntax.definition import EnclosedSyntaxDefinition


class FileLikeBlockRenderer(CaptionedBlockRenderer):
    @abstractmethod
    def _get_syntax(self) -> EnclosedSyntaxDefinition:
        pass

    @abstractmethod
    def _get_file_data(
        self, block: Block
    ) -> ExternalFileWithCaption | NotionHostedFileWithCaption | None:
        pass

    @override
    async def _render_main_content(self, block: Block) -> str:
        url = self._extract_url(block)

        if not url:
            return ""

        syntax = self._get_syntax()
        return f"{syntax.start_delimiter}{url}{syntax.end_delimiter}"

    def _extract_url(self, block: Block) -> str:
        file_data = self._get_file_data(block)

        if not file_data:
            return ""

        if isinstance(file_data, ExternalFileWithCaption):
            return file_data.external.url or ""
        elif isinstance(file_data, NotionHostedFileWithCaption):
            return file_data.file.url or ""

        return ""



from typing import override

from notionary.exceptions.file_upload import FilenameTooLongError
from notionary.file_upload.config.config import FileUploadConfig
from notionary.file_upload.validation.port import FileUploadValidator


class FileNameLengthValidator(FileUploadValidator):
    def __init__(
        self, filename: str, file_upload_config: FileUploadConfig | None = None
    ) -> None:
        self._filename = filename

        file_upload_config = file_upload_config or FileUploadConfig()
        self._max_filename_bytes = file_upload_config.MAX_FILENAME_BYTES

    @override
    async def validate(self) -> None:
        filename_bytes = len(self._filename.encode("utf-8"))
        if filename_bytes > self._max_filename_bytes:
            raise FilenameTooLongError(
                filename=self._filename,
                filename_bytes=filename_bytes,
                max_filename_bytes=self._max_filename_bytes,
            )



from typing import override

from notionary.blocks.schemas import Block
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer
from notionary.utils.mixins.logging import LoggingMixin


class FallbackRenderer(BlockRenderer, LoggingMixin):
    @override
    def _can_handle(self, block: Block) -> bool:
        return True

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        block_type = context.block.type.value if context.block.type else "unknown"
        self.logger.warning(f"No handler found for block type: {block_type}")

        fallback_message = f"[Unsupported block type: {block_type}]"

        if context.indent_level > 0:
            fallback_message = context.indent_text(fallback_message)

        context.markdown_result = fallback_message



from typing import Self, cast

from notionary.user.base import BaseUser
from notionary.user.schemas import PersonUserResponseDto, UserResponseDto, UserType


class PersonUser(BaseUser):
    def __init__(
        self,
        id: str,
        name: str,
        avatar_url: str,
        email: str,
    ) -> None:
        super().__init__(id=id, name=name, avatar_url=avatar_url)
        self._email = email

    @classmethod
    def _get_expected_user_type(cls) -> UserType:
        return UserType.PERSON

    @classmethod
    def from_dto(cls, user_dto: UserResponseDto) -> Self:
        person_dto = cast(PersonUserResponseDto, user_dto)
        return cls(
            id=person_dto.id,
            name=person_dto.name or "",
            avatar_url=person_dto.avatar_url,
            email=person_dto.person.email or "",
        )

    @property
    def name(self) -> str:
        return self._name or ""

    @property
    def email(self) -> str:
        return self._email

    def __repr__(self) -> str:
        return (
            f"PersonUser(id={self._id!r}, name={self._name!r}, email={self._email!r})"
        )



from abc import ABC, abstractmethod

from notionary.blocks.schemas import BlockCreatePayload


class PostProcessor(ABC):
    @abstractmethod
    def process(self, blocks: list[BlockCreatePayload]) -> list[BlockCreatePayload]:
        pass



from pydantic import BaseModel

from notionary.page.properties.schemas import AnyPageProperty
from notionary.shared.entity.schemas import EntityResponseDto


class NotionPageDto(EntityResponseDto):
    archived: bool
    properties: dict[str, AnyPageProperty]


class PgePropertiesUpdateDto(BaseModel):
    properties: dict[str, AnyPageProperty]



from typing import override

from notionary.blocks.enums import BlockColor
from notionary.blocks.rich_text.markdown_rich_text_converter import (
    MarkdownRichTextConverter,
)
from notionary.blocks.schemas import CreateParagraphBlock, CreateParagraphData
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)


class ParagraphParser(LineParser):
    def __init__(self, rich_text_converter: MarkdownRichTextConverter) -> None:
        super().__init__()
        self._rich_text_converter = rich_text_converter

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False
        return bool(context.line)

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        block = await self._create_paragraph_block(context.line)
        if block:
            context.result_blocks.append(block)

    async def _create_paragraph_block(self, text: str) -> CreateParagraphBlock | None:
        if not text:
            return None

        rich_text = await self._rich_text_converter.to_rich_text(text)
        paragraph_content = CreateParagraphData(
            rich_text=rich_text, color=BlockColor.DEFAULT
        )
        return CreateParagraphBlock(paragraph=paragraph_content)



from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.page.content.renderer.renderers import (
    AudioRenderer,
    BlockRenderer,
    BookmarkRenderer,
    BreadcrumbRenderer,
    BulletedListRenderer,
    CalloutRenderer,
    CodeRenderer,
    ColumnListRenderer,
    ColumnRenderer,
    DividerRenderer,
    EmbedRenderer,
    EquationRenderer,
    FallbackRenderer,
    FileRenderer,
    HeadingRenderer,
    ImageRenderer,
    NumberedListRenderer,
    ParagraphRenderer,
    PdfRenderer,
    QuoteRenderer,
    TableOfContentsRenderer,
    TableRenderer,
    TableRowHandler,
    TodoRenderer,
    ToggleRenderer,
    VideoRenderer,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class RendererChainFactory:
    def __init__(
        self,
        rich_text_markdown_converter: RichTextToMarkdownConverter | None = None,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
    ) -> None:
        self._rich_text_markdown_converter = (
            rich_text_markdown_converter or RichTextToMarkdownConverter()
        )
        self._syntax_registry = syntax_registry or SyntaxDefinitionRegistry()

    def create(self) -> BlockRenderer:
        # Strukturelle Blocks
        toggle_handler = self._create_toggle_renderer()
        heading_handler = self._create_heading_renderer()

        # Content Blocks
        callout_handler = self._create_callout_renderer()
        code_handler = self._create_code_renderer()
        quote_handler = self._create_quote_renderer()
        todo_handler = self._create_todo_renderer()
        bulleted_list_handler = self._create_bulleted_list_renderer()

        # Layout Blocks
        divider_handler = self._create_divider_renderer()
        column_list_handler = self._create_column_list_renderer()
        column_handler = self._create_column_renderer()
        numbered_list_handler = self._create_numbered_list_renderer()

        # Media Blocks
        bookmark_handler = self._create_bookmark_renderer()
        image_handler = self._create_image_renderer()
        video_handler = self._create_video_renderer()
        audio_handler = self._create_audio_renderer()
        file_handler = self._create_file_renderer()
        pdf_handler = self._create_pdf_renderer()
        embed_handler = self._create_embed_renderer()

        # Special Blocks
        equation_handler = self._create_equation_renderer()
        table_of_contents_handler = self._create_table_of_contents_renderer()
        breadcrumb_handler = self._create_breadcrumb_renderer()
        table_handler = self._create_table_renderer()
        table_row_handler = self._create_table_row_handler()

        # Standard & Fallback
        paragraph_handler = self._create_paragraph_renderer()
        fallback_handler = self._create_fallback_renderer()

        # Chain verketten - most specific first, fallback last
        (
            toggle_handler.set_next(heading_handler)
            .set_next(callout_handler)
            .set_next(code_handler)
            .set_next(quote_handler)
            .set_next(todo_handler)
            .set_next(bulleted_list_handler)
            .set_next(divider_handler)
            .set_next(column_list_handler)
            .set_next(column_handler)
            .set_next(numbered_list_handler)
            .set_next(bookmark_handler)
            .set_next(image_handler)
            .set_next(video_handler)
            .set_next(audio_handler)
            .set_next(file_handler)
            .set_next(pdf_handler)
            .set_next(embed_handler)
            .set_next(equation_handler)
            .set_next(table_of_contents_handler)
            .set_next(breadcrumb_handler)
            .set_next(table_handler)
            .set_next(table_row_handler)
            .set_next(paragraph_handler)
            .set_next(fallback_handler)
        )

        return toggle_handler

    # Renderer Creation Methods
    def _create_toggle_renderer(self) -> ToggleRenderer:
        return ToggleRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_heading_renderer(self) -> HeadingRenderer:
        return HeadingRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_callout_renderer(self) -> CalloutRenderer:
        return CalloutRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_code_renderer(self) -> CodeRenderer:
        return CodeRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_quote_renderer(self) -> QuoteRenderer:
        return QuoteRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_todo_renderer(self) -> TodoRenderer:
        return TodoRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_bulleted_list_renderer(self) -> BulletedListRenderer:
        return BulletedListRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_divider_renderer(self) -> DividerRenderer:
        return DividerRenderer(syntax_registry=self._syntax_registry)

    def _create_column_list_renderer(self) -> ColumnListRenderer:
        return ColumnListRenderer(syntax_registry=self._syntax_registry)

    def _create_column_renderer(self) -> ColumnRenderer:
        return ColumnRenderer(syntax_registry=self._syntax_registry)

    def _create_numbered_list_renderer(self) -> NumberedListRenderer:
        return NumberedListRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_bookmark_renderer(self) -> BookmarkRenderer:
        return BookmarkRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_image_renderer(self) -> ImageRenderer:
        return ImageRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_video_renderer(self) -> VideoRenderer:
        return VideoRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_audio_renderer(self) -> AudioRenderer:
        return AudioRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_file_renderer(self) -> FileRenderer:
        return FileRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_pdf_renderer(self) -> PdfRenderer:
        return PdfRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_embed_renderer(self) -> EmbedRenderer:
        return EmbedRenderer(
            syntax_registry=self._syntax_registry,
            rich_text_markdown_converter=self._rich_text_markdown_converter,
        )

    def _create_equation_renderer(self) -> EquationRenderer:
        return EquationRenderer(syntax_registry=self._syntax_registry)

    def _create_table_of_contents_renderer(self) -> TableOfContentsRenderer:
        return TableOfContentsRenderer(syntax_registry=self._syntax_registry)

    def _create_breadcrumb_renderer(self) -> BreadcrumbRenderer:
        return BreadcrumbRenderer(syntax_registry=self._syntax_registry)

    def _create_table_renderer(self) -> TableRenderer:
        return TableRenderer(
            rich_text_markdown_converter=self._rich_text_markdown_converter
        )

    def _create_table_row_handler(self) -> TableRowHandler:
        return TableRowHandler()

    def _create_paragraph_renderer(self) -> ParagraphRenderer:
        return ParagraphRenderer(
            rich_text_markdown_converter=self._rich_text_markdown_converter
        )

    def _create_fallback_renderer(self) -> FallbackRenderer:
        return FallbackRenderer()



from ..context import BlockParsingContext, ParentBlockContext
from .audio import AudioParser
from .base import LineParser
from .bookmark import BookmarkParser
from .breadcrumb import BreadcrumbParser
from .bulleted_list import BulletedListParser
from .callout import CalloutParser
from .caption import CaptionParser
from .code import CodeParser
from .column import ColumnParser
from .column_list import ColumnListParser
from .divider import DividerParser
from .embed import EmbedParser
from .equation import EquationParser
from .file import FileParser
from .heading import HeadingParser
from .image import ImageParser
from .numbered_list import NumberedListParser
from .paragraph import ParagraphParser
from .pdf import PdfParser
from .quote import QuoteParser
from .space import SpaceParser
from .table import TableParser
from .table_of_contents import TableOfContentsParser
from .todo import TodoParser
from .toggle import ToggleParser
from .video import VideoParser

__all__ = [
    "AudioParser",
    "BlockParsingContext",
    "BookmarkParser",
    "BreadcrumbParser",
    "BulletedListParser",
    "CalloutParser",
    "CaptionParser",
    "CodeParser",
    "ColumnListParser",
    "ColumnParser",
    "DividerParser",
    "EmbedParser",
    "EquationParser",
    "FileParser",
    "HeadingParser",
    "ImageParser",
    "LineParser",
    "NumberedListParser",
    "ParagraphParser",
    "ParentBlockContext",
    "PdfParser",
    "QuoteParser",
    "SpaceParser",
    "TableOfContentsParser",
    "TableParser",
    "TodoParser",
    "ToggleParser",
    "VideoParser",
]



from .audio import AudioMarkdownNode
from .base import MarkdownNode
from .bookmark import BookmarkMarkdownNode
from .breadcrumb import BreadcrumbMarkdownNode
from .bulleted_list import BulletedListMarkdownNode
from .callout import CalloutMarkdownNode
from .code import CodeMarkdownNode
from .columns import ColumnListMarkdownNode, ColumnMarkdownNode
from .divider import DividerMarkdownNode
from .embed import EmbedMarkdownNode
from .equation import EquationMarkdownNode
from .file import FileMarkdownNode
from .heading import HeadingMarkdownNode
from .image import ImageMarkdownNode
from .numbered_list import NumberedListMarkdownNode
from .paragraph import ParagraphMarkdownNode
from .pdf import PdfMarkdownNode
from .quote import QuoteMarkdownNode
from .space import SpaceMarkdownNode
from .table import TableMarkdownNode
from .table_of_contents import TableOfContentsMarkdownNode
from .todo import TodoMarkdownNode
from .toggle import ToggleMarkdownNode
from .video import VideoMarkdownNode

__all__ = [
    "AudioMarkdownNode",
    "BookmarkMarkdownNode",
    "BreadcrumbMarkdownNode",
    "BulletedListMarkdownNode",
    "CalloutMarkdownNode",
    "CodeMarkdownNode",
    "ColumnListMarkdownNode",
    "ColumnMarkdownNode",
    "DividerMarkdownNode",
    "EmbedMarkdownNode",
    "EquationMarkdownNode",
    "FileMarkdownNode",
    "HeadingMarkdownNode",
    "ImageMarkdownNode",
    "MarkdownNode",
    "NumberedListMarkdownNode",
    "ParagraphMarkdownNode",
    "PdfMarkdownNode",
    "QuoteMarkdownNode",
    "SpaceMarkdownNode",
    "TableMarkdownNode",
    "TableOfContentsMarkdownNode",
    "TodoMarkdownNode",
    "ToggleMarkdownNode",
    "VideoMarkdownNode",
]



from typing import override

from notionary.blocks.rich_text.markdown_rich_text_converter import (
    MarkdownRichTextConverter,
)
from notionary.blocks.schemas import BlockCreatePayload
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class CaptionParser(LineParser):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry,
        rich_text_converter: MarkdownRichTextConverter,
    ) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_caption_syntax()
        self._rich_text_converter = rich_text_converter

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False

        if not self._syntax.regex_pattern.match(context.line):
            return False

        if not context.result_blocks:
            return False

        previous_block = context.result_blocks[-1]
        return self._block_supports_caption(previous_block)

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        caption_match = self._syntax.regex_pattern.match(context.line)
        if not caption_match:
            return

        caption_text = caption_match.group(1)
        caption_rich_text = await self._rich_text_converter.to_rich_text(caption_text)

        previous_block = context.result_blocks[-1]
        self._attach_caption_to_block(previous_block, caption_rich_text)

    def _block_supports_caption(self, block: BlockCreatePayload) -> bool:
        block_data = getattr(block, block.type.value, None)
        if block_data is None:
            return False
        return hasattr(block_data, "caption")

    def _attach_caption_to_block(
        self, block: BlockCreatePayload, caption_rich_text: list
    ) -> None:
        block_data = getattr(block, block.type.value)
        if hasattr(block_data, "caption"):
            block_data.caption = caption_rich_text



import re
from typing import override
from urllib.parse import urlparse

from notionary.blocks.enums import VideoFileType
from notionary.exceptions import UnsupportedVideoFormatError
from notionary.page.content.parser.pre_processsing.handlers.port import PreProcessor
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry
from notionary.utils.decorators import time_execution_sync
from notionary.utils.mixins.logging import LoggingMixin


class VideoFormatPreProcessor(PreProcessor, LoggingMixin):
    YOUTUBE_WATCH_PATTERN = re.compile(
        r"^https?://(?:www\.)?youtube\.com/watch\?.*v=[\w-]+", re.IGNORECASE
    )
    YOUTUBE_EMBED_PATTERN = re.compile(
        r"^https?://(?:www\.)?youtube\.com/embed/[\w-]+", re.IGNORECASE
    )

    def __init__(self, syntax_registry: SyntaxDefinitionRegistry | None = None) -> None:
        super().__init__()
        self._syntax_registry = syntax_registry or SyntaxDefinitionRegistry()
        self._video_syntax = self._syntax_registry.get_video_syntax()

    @override
    @time_execution_sync()
    def process(self, markdown_text: str) -> str:
        lines = markdown_text.split("\n")
        validated_lines = [self._validate_or_reject_line(line) for line in lines]
        return "\n".join(validated_lines)

    def _validate_or_reject_line(self, line: str) -> str:
        if not self._contains_video_block(line):
            return line

        url = self._extract_url_from_video_block(line)

        if self._is_supported_video_url(url):
            return line

        supported_formats = list(VideoFileType.get_all_extensions())
        raise UnsupportedVideoFormatError(url, supported_formats)

    def _contains_video_block(self, line: str) -> bool:
        return self._video_syntax.regex_pattern.search(line) is not None

    def _extract_url_from_video_block(self, line: str) -> str:
        match = self._video_syntax.regex_pattern.search(line)
        return match.group(1).strip() if match else ""

    def _is_supported_video_url(self, url: str) -> bool:
        return (
            self._is_youtube_video(url)
            or self._has_valid_video_extension(url)
            or self._url_path_has_valid_extension(url)
        )

    def _is_youtube_video(self, url: str) -> bool:
        return bool(
            self.YOUTUBE_WATCH_PATTERN.match(url)
            or self.YOUTUBE_EMBED_PATTERN.match(url)
        )

    def _has_valid_video_extension(self, url: str) -> bool:
        return VideoFileType.is_valid_extension(url)

    def _url_path_has_valid_extension(self, url: str) -> bool:
        try:
            parsed_url = urlparse(url)
            return VideoFileType.is_valid_extension(parsed_url.path.lower())
        except Exception:
            return False



from .file_exists import FileExistsValidator
from .file_extension import FileExtensionValidator
from .file_name_length import FileNameLengthValidator
from .upload_limit import FileUploadLimitValidator

__all__ = [
    "FileExistsValidator",
    "FileExtensionValidator",
    "FileNameLengthValidator",
    "FileUploadLimitValidator",
]



from collections.abc import Callable
from typing import Self

from notionary.data_source.properties.schemas import DataSourceProperty
from notionary.data_source.query.schema import (
    ArrayOperator,
    BooleanOperator,
    CompoundFilter,
    DataSourceQueryParams,
    DateOperator,
    FieldType,
    FilterCondition,
    InternalFilterCondition,
    LogicalOperator,
    NotionFilter,
    NotionSort,
    NumberOperator,
    Operator,
    OrGroupMarker,
    PropertyFilter,
    PropertySort,
    SortDirection,
    StringOperator,
    TimestampSort,
    TimestampType,
)
from notionary.data_source.query.validator import QueryValidator
from notionary.exceptions.data_source.properties import DataSourcePropertyNotFound
from notionary.utils.date import parse_date


class DataSourceQueryBuilder:
    def __init__(
        self,
        properties: dict[str, DataSourceProperty],
        query_validator: QueryValidator | None = None,
        date_parser: Callable[[str], str] = parse_date,
    ) -> None:
        self._properties = properties
        self._query_validator = query_validator or QueryValidator()
        self._date_parser = date_parser

        self._filters: list[InternalFilterCondition] = []
        self._sorts: list[NotionSort] = []
        self._current_property: str | None = None
        self._negate_next = False
        self._or_group: list[FilterCondition] | None = None
        self._page_size: int | None = None
        self._total_results_limit: int | None = None

    def where(self, property_name: str) -> Self:
        self._finalize_current_or_group()
        self._ensure_property_exists(property_name)
        self._select_property_without_negation(property_name)
        return self

    def where_not(self, property_name: str) -> Self:
        self._finalize_current_or_group()
        self._ensure_property_exists(property_name)
        self._select_property_with_negation(property_name)
        return self

    def and_where(self, property_name: str) -> Self:
        return self.where(property_name)

    def and_where_not(self, property_name: str) -> Self:
        return self.where_not(property_name)

    def or_where(self, property_name: str) -> Self:
        self._ensure_or_group_exists()
        self._ensure_property_exists(property_name)
        self._select_property_without_negation(property_name)
        return self

    def or_where_not(self, property_name: str) -> Self:
        self._ensure_or_group_exists()
        self._ensure_property_exists(property_name)
        self._select_property_with_negation(property_name)
        return self

    def equals(self, value: str | int | float) -> Self:
        return self._add_filter(StringOperator.EQUALS, value)

    def does_not_equal(self, value: str | int | float) -> Self:
        return self._add_filter(StringOperator.DOES_NOT_EQUAL, value)

    def contains(self, value: str) -> Self:
        return self._add_filter(StringOperator.CONTAINS, value)

    def does_not_contain(self, value: str) -> Self:
        return self._add_filter(StringOperator.DOES_NOT_CONTAIN, value)

    def starts_with(self, value: str) -> Self:
        return self._add_filter(StringOperator.STARTS_WITH, value)

    def ends_with(self, value: str) -> Self:
        return self._add_filter(StringOperator.ENDS_WITH, value)

    def is_empty(self) -> Self:
        return self._add_filter(StringOperator.IS_EMPTY, None)

    def is_not_empty(self) -> Self:
        return self._add_filter(StringOperator.IS_NOT_EMPTY, None)

    def greater_than(self, value: float | int) -> Self:
        return self._add_filter(NumberOperator.GREATER_THAN, value)

    def greater_than_or_equal_to(self, value: float | int) -> Self:
        return self._add_filter(NumberOperator.GREATER_THAN_OR_EQUAL_TO, value)

    def less_than(self, value: float | int) -> Self:
        return self._add_filter(NumberOperator.LESS_THAN, value)

    def less_than_or_equal_to(self, value: float | int) -> Self:
        return self._add_filter(NumberOperator.LESS_THAN_OR_EQUAL_TO, value)

    def is_true(self) -> Self:
        return self._add_filter(BooleanOperator.IS_TRUE, None)

    def is_false(self) -> Self:
        return self._add_filter(BooleanOperator.IS_FALSE, None)

    def before(self, date: str) -> Self:
        parsed_date = self._date_parser(date)
        return self._add_filter(DateOperator.BEFORE, parsed_date)

    def after(self, date: str) -> Self:
        parsed_date = self._date_parser(date)
        return self._add_filter(DateOperator.AFTER, parsed_date)

    def on_or_before(self, date: str) -> Self:
        parsed_date = self._date_parser(date)
        return self._add_filter(DateOperator.ON_OR_BEFORE, parsed_date)

    def on_or_after(self, date: str) -> Self:
        parsed_date = self._date_parser(date)
        return self._add_filter(DateOperator.ON_OR_AFTER, parsed_date)

    def array_contains(self, value: str) -> Self:
        return self._add_filter(ArrayOperator.CONTAINS, value)

    def array_does_not_contain(self, value: str) -> Self:
        return self._add_filter(ArrayOperator.DOES_NOT_CONTAIN, value)

    def array_is_empty(self) -> Self:
        return self._add_filter(ArrayOperator.IS_EMPTY, None)

    def array_is_not_empty(self) -> Self:
        return self._add_filter(ArrayOperator.IS_NOT_EMPTY, None)

    def relation_contains(self, uuid: str) -> Self:
        return self._add_filter(ArrayOperator.CONTAINS, uuid)

    def relation_is_empty(self) -> Self:
        return self._add_filter(ArrayOperator.IS_EMPTY, None)

    def people_contains(self, uuid: str) -> Self:
        return self._add_filter(ArrayOperator.CONTAINS, uuid)

    def people_is_empty(self) -> Self:
        return self._add_filter(ArrayOperator.IS_EMPTY, None)

    def order_by(
        self, property_name: str, direction: SortDirection = SortDirection.ASCENDING
    ) -> Self:
        self._ensure_property_exists(property_name)
        sort = PropertySort(property=property_name, direction=direction)
        self._sorts.append(sort)
        return self

    def order_by_property_name_ascending(self, property_name: str) -> Self:
        return self.order_by(property_name, SortDirection.ASCENDING)

    def order_by_property_name_descending(self, property_name: str) -> Self:
        return self.order_by(property_name, SortDirection.DESCENDING)

    def order_by_created_time_ascending(self) -> Self:
        return self._order_by_created_time(SortDirection.ASCENDING)

    def order_by_created_time_descending(self) -> Self:
        return self._order_by_created_time(SortDirection.DESCENDING)

    def _order_by_created_time(
        self, direction: SortDirection = SortDirection.DESCENDING
    ) -> Self:
        sort = TimestampSort(timestamp=TimestampType.CREATED_TIME, direction=direction)
        self._sorts.append(sort)
        return self

    def order_by_last_edited_time_ascending(self) -> Self:
        return self._order_by_last_edited_time(SortDirection.ASCENDING)

    def order_by_last_edited_time_descending(self) -> Self:
        return self._order_by_last_edited_time(SortDirection.DESCENDING)

    def _order_by_last_edited_time(
        self, direction: SortDirection = SortDirection.DESCENDING
    ) -> Self:
        sort = TimestampSort(
            timestamp=TimestampType.LAST_EDITED_TIME, direction=direction
        )
        self._sorts.append(sort)
        return self

    def total_results_limit(self, total_result_limit: int) -> Self:
        if total_result_limit < 1:
            raise ValueError("Limit must be at least 1")
        self._total_results_limit = total_result_limit
        return self

    def page_size(self, page_size: int) -> Self:
        if page_size < 1:
            raise ValueError("Page size must be at least 1")
        self._page_size = page_size
        return self

    def build(self) -> DataSourceQueryParams:
        self._finalize_current_or_group()
        notion_filter = self._create_notion_filter_if_needed()
        sorts = self._create_sorts_if_needed()
        return DataSourceQueryParams(
            filter=notion_filter,
            sorts=sorts,
            page_size=self._page_size,
            total_results_limit=self._total_results_limit,
        )

    def _select_property_without_negation(self, property_name: str) -> None:
        self._current_property = property_name
        self._negate_next = False

    def _select_property_with_negation(self, property_name: str) -> None:
        self._current_property = property_name
        self._negate_next = True

    def _ensure_property_exists(self, property_name: str) -> None:
        if self._has_no_properties():
            return

        if self._property_is_unknown(property_name):
            self._raise_property_not_found_error(property_name)

    def _has_no_properties(self) -> bool:
        return not self._properties

    def _property_is_unknown(self, property_name: str) -> bool:
        return property_name not in self._properties

    def _raise_property_not_found_error(self, property_name: str) -> None:
        available_properties = list(self._properties.keys())
        raise DataSourcePropertyNotFound(
            property_name=property_name,
            available_properties=available_properties,
        )

    def _ensure_or_group_exists(self) -> None:
        if self._or_group_is_not_active():
            self._start_new_or_group()

    def _or_group_is_not_active(self) -> bool:
        return self._or_group is None

    def _start_new_or_group(self) -> None:
        self._or_group = []
        self._move_last_filter_to_or_group()

    def _move_last_filter_to_or_group(self) -> None:
        if self._has_no_filters():
            return

        last_filter = self._filters.pop()
        if self._is_regular_filter_condition(last_filter):
            self._or_group.append(last_filter)

    def _has_no_filters(self) -> bool:
        return not self._filters

    def _is_regular_filter_condition(
        self, filter_item: InternalFilterCondition
    ) -> bool:
        return isinstance(filter_item, FilterCondition)

    def _finalize_current_or_group(self) -> None:
        if self._should_finalize_or_group():
            self._convert_or_group_to_marker()
            self._clear_or_group()

    def _should_finalize_or_group(self) -> bool:
        return self._or_group is not None and len(self._or_group) > 0

    def _convert_or_group_to_marker(self) -> None:
        or_marker = OrGroupMarker(conditions=self._or_group)
        self._filters.append(or_marker)

    def _clear_or_group(self) -> None:
        self._or_group = None

    def _add_filter(
        self,
        operator: StringOperator
        | NumberOperator
        | BooleanOperator
        | DateOperator
        | ArrayOperator,
        value: str | int | float | list[str | int | float] | None,
    ) -> Self:
        self._ensure_property_is_selected()
        self._validate_operator_for_current_property(operator)
        final_operator = self._apply_negation_if_needed(operator)
        filter_condition = self._create_filter_condition(final_operator, value)
        self._store_filter_condition(filter_condition)
        self._reset_current_property()
        return self

    def _validate_operator_for_current_property(self, operator: Operator) -> None:
        if not self._current_property or not self._properties:
            return

        property_obj = self._properties.get(self._current_property)
        if property_obj:
            self._query_validator.validate_operator_for_property(
                self._current_property, property_obj, operator
            )
        return self

    def _ensure_property_is_selected(self) -> None:
        if self._no_property_is_selected():
            raise ValueError("No property selected. Use .where(property_name) first.")

    def _no_property_is_selected(self) -> bool:
        return self._current_property is None

    def _apply_negation_if_needed(
        self,
        operator: StringOperator
        | NumberOperator
        | BooleanOperator
        | DateOperator
        | ArrayOperator,
    ) -> (
        StringOperator | NumberOperator | BooleanOperator | DateOperator | ArrayOperator
    ):
        if not self._negate_next:
            return operator

        negated_operator = self._negate_operator(operator)
        self._negate_next = False
        return negated_operator

    def _create_filter_condition(
        self,
        operator: StringOperator
        | NumberOperator
        | BooleanOperator
        | DateOperator
        | ArrayOperator,
        value: str | int | float | list[str | int | float] | None,
    ) -> FilterCondition:
        field_type = self._determine_field_type_from_operator(operator)
        return FilterCondition(
            field=self._current_property,
            field_type=field_type,
            operator=operator,
            value=value,
        )

    def _store_filter_condition(self, filter_condition: FilterCondition) -> None:
        if self._or_group_is_active():
            self._or_group.append(filter_condition)
        else:
            self._filters.append(filter_condition)

    def _or_group_is_active(self) -> bool:
        return self._or_group is not None

    def _reset_current_property(self) -> None:
        self._current_property = None

    def _create_notion_filter_if_needed(self) -> NotionFilter | None:
        if self._has_no_filters():
            return None
        return self._build_notion_filter()

    def _create_sorts_if_needed(self) -> list[NotionSort] | None:
        if not self._sorts:
            return None
        return self._sorts

    def _build_notion_filter(self) -> NotionFilter:
        if self._has_single_filter():
            return self._build_single_filter()
        return self._build_compound_and_filter()

    def _has_single_filter(self) -> bool:
        return len(self._filters) == 1

    def _build_single_filter(self) -> NotionFilter:
        return self._build_filter(self._filters[0])

    def _build_compound_and_filter(self) -> CompoundFilter:
        property_filters = [self._build_filter(f) for f in self._filters]
        return CompoundFilter(operator=LogicalOperator.AND, filters=property_filters)

    def _build_filter(
        self, condition: InternalFilterCondition
    ) -> PropertyFilter | CompoundFilter:
        if isinstance(condition, OrGroupMarker):
            return self._build_or_compound_filter(condition)
        return self._build_property_filter(condition)

    def _build_or_compound_filter(self, or_marker: OrGroupMarker) -> CompoundFilter:
        property_filters = [
            self._build_property_filter(c) for c in or_marker.conditions
        ]
        return CompoundFilter(operator=LogicalOperator.OR, filters=property_filters)

    def _build_property_filter(self, condition: FilterCondition) -> PropertyFilter:
        property_definition = self._get_property_definition(condition.field)
        return PropertyFilter(
            property=condition.field,
            property_type=property_definition.type,
            operator=condition.operator,
            value=condition.value,
        )

    def _get_property_definition(self, property_name: str) -> DataSourceProperty:
        property_definition = self._properties.get(property_name)
        if property_definition is None:
            self._raise_property_not_found_error(property_name)
        return property_definition

    def _negate_operator(
        self,
        operator: StringOperator
        | NumberOperator
        | BooleanOperator
        | DateOperator
        | ArrayOperator,
    ) -> (
        StringOperator | NumberOperator | BooleanOperator | DateOperator | ArrayOperator
    ):
        negation_map = {
            StringOperator.EQUALS: StringOperator.DOES_NOT_EQUAL,
            StringOperator.DOES_NOT_EQUAL: StringOperator.EQUALS,
            StringOperator.CONTAINS: StringOperator.DOES_NOT_CONTAIN,
            StringOperator.DOES_NOT_CONTAIN: StringOperator.CONTAINS,
            StringOperator.IS_EMPTY: StringOperator.IS_NOT_EMPTY,
            StringOperator.IS_NOT_EMPTY: StringOperator.IS_EMPTY,
            NumberOperator.EQUALS: NumberOperator.DOES_NOT_EQUAL,
            NumberOperator.DOES_NOT_EQUAL: NumberOperator.EQUALS,
            NumberOperator.GREATER_THAN: NumberOperator.LESS_THAN_OR_EQUAL_TO,
            NumberOperator.GREATER_THAN_OR_EQUAL_TO: NumberOperator.LESS_THAN,
            NumberOperator.LESS_THAN: NumberOperator.GREATER_THAN_OR_EQUAL_TO,
            NumberOperator.LESS_THAN_OR_EQUAL_TO: NumberOperator.GREATER_THAN,
            NumberOperator.IS_EMPTY: NumberOperator.IS_NOT_EMPTY,
            NumberOperator.IS_NOT_EMPTY: NumberOperator.IS_EMPTY,
            BooleanOperator.IS_TRUE: BooleanOperator.IS_FALSE,
            BooleanOperator.IS_FALSE: BooleanOperator.IS_TRUE,
            DateOperator.BEFORE: DateOperator.ON_OR_AFTER,
            DateOperator.AFTER: DateOperator.ON_OR_BEFORE,
            DateOperator.ON_OR_BEFORE: DateOperator.AFTER,
            DateOperator.ON_OR_AFTER: DateOperator.BEFORE,
            DateOperator.IS_EMPTY: DateOperator.IS_NOT_EMPTY,
            DateOperator.IS_NOT_EMPTY: DateOperator.IS_EMPTY,
            ArrayOperator.CONTAINS: ArrayOperator.DOES_NOT_CONTAIN,
            ArrayOperator.DOES_NOT_CONTAIN: ArrayOperator.CONTAINS,
            ArrayOperator.IS_EMPTY: ArrayOperator.IS_NOT_EMPTY,
            ArrayOperator.IS_NOT_EMPTY: ArrayOperator.IS_EMPTY,
        }

        if operator not in negation_map:
            self._raise_operator_cannot_be_negated_error(operator)

        return negation_map[operator]

    def _raise_operator_cannot_be_negated_error(
        self,
        operator: StringOperator
        | NumberOperator
        | BooleanOperator
        | DateOperator
        | ArrayOperator,
    ) -> None:
        raise ValueError(
            f"Operator '{operator}' cannot be negated. This should not happen - please report this issue."
        )

    def _determine_field_type_from_operator(
        self,
        operator: StringOperator
        | NumberOperator
        | BooleanOperator
        | DateOperator
        | ArrayOperator,
    ) -> FieldType:
        if isinstance(operator, StringOperator):
            return FieldType.STRING
        if isinstance(operator, NumberOperator):
            return FieldType.NUMBER
        if isinstance(operator, BooleanOperator):
            return FieldType.BOOLEAN
        if isinstance(operator, DateOperator):
            return FieldType.DATE
        if isinstance(operator, ArrayOperator):
            return FieldType.ARRAY
        return FieldType.STRING



from typing import override

from notionary.page.content.parser.pre_processsing.handlers.port import PreProcessor
from notionary.utils.decorators import time_execution_sync


class WhitespacePreProcessor(PreProcessor):
    @override
    @time_execution_sync()
    def process(self, markdown_text: str) -> str:
        if not markdown_text:
            return ""

        lines = markdown_text.split("\n")
        processed_lines = []
        code_block_lines = []
        non_code_lines = []
        in_code_block = False

        for line in lines:
            if self._is_code_fence(line):
                if in_code_block:
                    # Format and add code block
                    processed_lines.extend(self._format_code_block(code_block_lines))
                    processed_lines.append("```")
                    code_block_lines = []
                    in_code_block = False
                else:
                    # Format accumulated non-code lines before starting code block
                    if non_code_lines:
                        processed_lines.extend(self._format_code_block(non_code_lines))
                        non_code_lines = []

                    language = self._extract_language(line)
                    processed_lines.append(f"```{language}")
                    in_code_block = True
            elif in_code_block:
                code_block_lines.append(line)
            else:
                non_code_lines.append(line)

        # Format remaining non-code lines at the end
        if non_code_lines:
            processed_lines.extend(self._format_code_block(non_code_lines))

        return "\n".join(processed_lines)

    def _is_code_fence(self, line: str) -> bool:
        return line.lstrip().startswith("```")

    def _extract_language(self, fence_line: str) -> str:
        return fence_line.lstrip().removeprefix("```").strip()

    def _format_code_block(self, lines: list[str]) -> list[str]:
        if not lines:
            return []

        non_empty_lines = [line for line in lines if line.strip()]
        if not non_empty_lines:
            return ["" for _ in lines]

        min_indent = min(self._count_leading_spaces(line) for line in non_empty_lines)

        if min_indent == 0:
            return lines

        return [self._remove_indent(line, min_indent) for line in lines]

    def _count_leading_spaces(self, line: str) -> int:
        return len(line) - len(line.lstrip())

    def _remove_indent(self, line: str, indent_size: int) -> str:
        if not line.strip():
            return ""
        return line[indent_size:]



import re
from enum import IntEnum
from typing import override

from notionary.page.content.renderer.post_processing.port import PostProcessor
from notionary.page.content.syntax.definition.grammar import MarkdownGrammar


class _NumberingStyle(IntEnum):
    NUMERIC = 0
    ALPHABETIC = 1
    ROMAN = 2


class _ListNumberingState:
    def __init__(self):
        self._counters_by_level: dict[int, int] = {}
        self._current_level = -1

    def advance_to_level(self, new_level: int) -> None:
        self._forget_deeper_levels_than(new_level)
        self._increment_counter_at_level(new_level)
        self._current_level = new_level

    def get_number_for_current_level(self) -> str:
        counter = self._counters_by_level[self._current_level]
        style = self._determine_numbering_style(self._current_level)
        return self._format_number(counter, style)

    def reset(self) -> None:
        self._counters_by_level.clear()
        self._current_level = -1

    def _forget_deeper_levels_than(self, level: int) -> None:
        self._counters_by_level = {
            existing_level: count
            for existing_level, count in self._counters_by_level.items()
            if existing_level <= level
        }

    def _increment_counter_at_level(self, level: int) -> None:
        self._counters_by_level[level] = self._counters_by_level.get(level, 0) + 1

    def _determine_numbering_style(self, nesting_level: int) -> _NumberingStyle:
        return _NumberingStyle(nesting_level % 3)

    def _format_number(self, counter: int, style: _NumberingStyle) -> str:
        if style == _NumberingStyle.NUMERIC:
            return str(counter)
        elif style == _NumberingStyle.ALPHABETIC:
            return self._to_alphabetic(counter)
        else:
            return self._to_roman(counter)

    def _to_alphabetic(self, number: int) -> str:
        result = ""
        number -= 1
        while number >= 0:
            result = chr(ord("a") + (number % 26)) + result
            number = number // 26 - 1
        return result

    def _to_roman(self, number: int) -> str:
        conversions = [
            (1000, "m"),
            (900, "cm"),
            (500, "d"),
            (400, "cd"),
            (100, "c"),
            (90, "xc"),
            (50, "l"),
            (40, "xl"),
            (10, "x"),
            (9, "ix"),
            (5, "v"),
            (4, "iv"),
            (1, "i"),
        ]

        result = ""
        for arabic, roman in conversions:
            while number >= arabic:
                result += roman
                number -= arabic
        return result


class NumberedListPlaceholderReplacerPostProcessor(PostProcessor):
    def __init__(self, markdown_grammar: MarkdownGrammar | None = None) -> None:
        self._markdown_grammar = markdown_grammar or MarkdownGrammar()
        self._spaces_per_nesting_level = self._markdown_grammar.spaces_per_nesting_level
        self._numbered_list_placeholder = (
            self._markdown_grammar.numbered_list_placeholder
        )

    @override
    def process(self, markdown_text: str) -> str:
        lines = markdown_text.splitlines()
        return self._convert_placeholder_lists_to_numbered_lists(lines)

    def _convert_placeholder_lists_to_numbered_lists(self, lines: list[str]) -> str:
        result = []
        list_state = _ListNumberingState()

        for line_index, line in enumerate(lines):
            if self._is_placeholder_list_item(line):
                numbered_line = self._convert_to_numbered_item(line, list_state)
                result.append(numbered_line)
            elif self._is_blank_between_list_items(lines, line_index, result):
                continue
            else:
                list_state.reset()
                result.append(line)

        return "\n".join(result)

    def _convert_to_numbered_item(self, line: str, state: _ListNumberingState) -> str:
        indentation = self._extract_indentation(line)
        content = self._extract_content(line)
        nesting_level = self._calculate_nesting_level(indentation)

        state.advance_to_level(nesting_level)
        number = state.get_number_for_current_level()

        return f"{indentation}{number}. {content}"

    def _calculate_nesting_level(self, indentation: str) -> int:
        return len(indentation) // self._spaces_per_nesting_level

    def _extract_indentation(self, line: str) -> str:
        match = re.match(rf"^(\s*){re.escape(self._numbered_list_placeholder)}\.", line)
        return match.group(1) if match else ""

    def _extract_content(self, line: str) -> str:
        match = re.match(
            rf"^\s*{re.escape(self._numbered_list_placeholder)}\.\s*(.*)", line
        )
        return match.group(1) if match else ""

    def _is_placeholder_list_item(self, line: str) -> bool:
        return bool(
            re.match(rf"^\s*{re.escape(self._numbered_list_placeholder)}\.", line)
        )

    def _is_blank_between_list_items(
        self, lines: list[str], current_index: int, processed_lines: list[str]
    ) -> bool:
        if not self._is_blank(lines[current_index]):
            return False

        previous_line_was_list_item = (
            processed_lines and self._looks_like_numbered_list_item(processed_lines[-1])
        )
        if not previous_line_was_list_item:
            return False

        next_line_is_list_item = current_index + 1 < len(
            lines
        ) and self._is_placeholder_list_item(lines[current_index + 1])
        return next_line_is_list_item

    def _is_blank(self, line: str) -> bool:
        return not line.strip()

    def _looks_like_numbered_list_item(self, line: str) -> bool:
        return bool(re.match(r"^\s*(\d+|[a-z]+|[ivxlcdm]+)\.\s+", line))



from typing import override

from notionary.blocks.schemas import (
    BlockColor,
    CreateTableOfContentsBlock,
    TableOfContentsData,
)
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class TableOfContentsParser(LineParser):
    def __init__(self, syntax_registry: SyntaxDefinitionRegistry) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_table_of_contents_syntax()

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False
        return self._is_toc(context.line)

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        block = self._create_toc_block()
        context.result_blocks.append(block)

    def _is_toc(self, line: str) -> bool:
        return self._syntax.regex_pattern.match(line) is not None

    def _create_toc_block(self) -> CreateTableOfContentsBlock:
        toc_data = TableOfContentsData(color=BlockColor.DEFAULT)
        return CreateTableOfContentsBlock(table_of_contents=toc_data)



from dataclasses import dataclass


@dataclass(frozen=True)
class SyntaxPromptData:
    element: str
    description: str
    is_multi_line: bool
    few_shot_examples: list[str]
    usage_notes: str
    supports_inline_rich_text: bool



"""
Handles request limits for rich texts (see https://developers.notion.com/reference/request-limits)
"""

from typing import Any, override

from notionary.blocks.rich_text.models import RichText, RichTextType
from notionary.blocks.schemas import BlockCreatePayload
from notionary.page.content.parser.post_processing.port import PostProcessor
from notionary.utils.mixins.logging import LoggingMixin

type _NestedBlockList = BlockCreatePayload | list["_NestedBlockList"]


class RichTextLengthTruncationPostProcessor(PostProcessor, LoggingMixin):
    NOTION_MAX_LENGTH = 2000
    ELLIPSIS = "..."

    def __init__(self, max_text_length: int = NOTION_MAX_LENGTH) -> None:
        self.max_text_length = max_text_length

    @override
    def process(self, blocks: list[BlockCreatePayload]) -> list[BlockCreatePayload]:
        if not blocks:
            return blocks

        flattened_blocks = self._flatten_blocks(blocks)
        return [self._process_block(block) for block in flattened_blocks]

    def _flatten_blocks(
        self, blocks: list[_NestedBlockList]
    ) -> list[BlockCreatePayload]:
        flattened: list[BlockCreatePayload] = []

        for item in blocks:
            if isinstance(item, list):
                flattened.extend(self._flatten_blocks(item))
            else:
                flattened.append(item)

        return flattened

    def _process_block(self, block: BlockCreatePayload) -> BlockCreatePayload:
        block_copy = block.model_copy(deep=True)
        content = self._get_block_content(block_copy)

        if content is not None:
            self._truncate_content(content)

        return block_copy

    def _get_block_content(self, block: BlockCreatePayload) -> Any | None:
        content = getattr(block, block.type.value, None)

        if content is None:
            return None

        if hasattr(content, "rich_text") or hasattr(content, "children"):
            return content

        return None

    def _truncate_content(self, content: object) -> None:
        self._truncate_rich_text_fields(content)
        self._truncate_children_recursively(content)

    def _truncate_rich_text_fields(self, content: object) -> None:
        if hasattr(content, "rich_text"):
            self._truncate_rich_text_list(content.rich_text)

        if hasattr(content, "caption"):
            self._truncate_rich_text_list(content.caption)

    def _truncate_children_recursively(self, content: object) -> None:
        if not hasattr(content, "children"):
            return

        children = getattr(content, "children", None)
        if not children:
            return

        for child in children:
            self._truncate_child_content(child)

    def _truncate_child_content(self, child: Any) -> None:
        child_content = self._get_block_content(child)
        if child_content:
            self._truncate_content(child_content)

    def _truncate_rich_text_list(self, rich_text_list: list[RichText]) -> None:
        for rich_text in rich_text_list:
            if self._should_truncate(rich_text):
                self._truncate_single_rich_text(rich_text)

    def _should_truncate(self, rich_text: RichText) -> bool:
        if not self._is_text_type(rich_text):
            return False

        return len(rich_text.text.content) > self.max_text_length

    def _truncate_single_rich_text(self, rich_text: RichText) -> None:
        original_length = len(rich_text.text.content)
        rich_text.text.content = self._create_truncated_text(rich_text.text.content)

        self.logger.warning(
            "Truncating text content from %d to %d characters",
            original_length,
            self.max_text_length,
        )

    def _create_truncated_text(self, content: str) -> str:
        cutoff = self.max_text_length - len(self.ELLIPSIS)
        return content[:cutoff] + self.ELLIPSIS

    def _is_text_type(self, rich_text: RichText) -> bool:
        return (
            rich_text.type == RichTextType.TEXT
            and rich_text.text is not None
            and rich_text.text.content
        )



from __future__ import annotations

from datetime import datetime
from enum import StrEnum
from typing import Literal

from pydantic import BaseModel, Field

from notionary.blocks.rich_text.models import RichText


class CommentParentType(StrEnum):
    PAGE_ID = "page_id"
    BLOCK_ID = "block_id"


class PageCommentParent(BaseModel):
    type: Literal[CommentParentType.PAGE_ID] = CommentParentType.PAGE_ID
    page_id: str


class BlockCommentParent(BaseModel):
    type: Literal[CommentParentType.BLOCK_ID] = CommentParentType.BLOCK_ID
    block_id: str


type CommentParent = PageCommentParent | BlockCommentParent


# ---------------------------
# Comment Attachment (Response/DTO)
# ---------------------------


class CommentAttachmentCategory(StrEnum):
    AUDIO = "audio"
    IMAGE = "image"
    PDF = "pdf"
    PRODUCTIVITY = "productivity"
    VIDEO = "video"


class FileWithExpiry(BaseModel):
    url: str
    expiry_time: datetime


class CommentAttachmentDto(BaseModel):
    category: CommentAttachmentCategory
    file: FileWithExpiry


# ---------------------------
# Comment Attachment (Request/Input)
# ---------------------------


class CommentAttachmentFileUploadType(StrEnum):
    FILE_UPLOAD = "file_upload"


class CommentAttachmentInput(BaseModel):
    file_upload_id: str
    type: Literal[CommentAttachmentFileUploadType.FILE_UPLOAD] = (
        CommentAttachmentFileUploadType.FILE_UPLOAD
    )


# ---------------------------
# Comment Display Name
# ---------------------------


class CommentDisplayNameType(StrEnum):
    INTEGRATION = "integration"
    USER = "user"
    CUSTOM = "custom"


class CustomDisplayName(BaseModel):
    name: str


class IntegrationDisplayName(BaseModel):
    type: Literal[CommentDisplayNameType.INTEGRATION] = (
        CommentDisplayNameType.INTEGRATION
    )


class UserDisplayName(BaseModel):
    type: Literal[CommentDisplayNameType.USER] = CommentDisplayNameType.USER


class CustomCommentDisplayName(BaseModel):
    type: Literal[CommentDisplayNameType.CUSTOM] = CommentDisplayNameType.CUSTOM
    custom: CustomDisplayName


type CommentDisplayNameInput = (
    IntegrationDisplayName | UserDisplayName | CustomCommentDisplayName
)


class CommentDisplayNameDto(BaseModel):
    type: CommentDisplayNameType
    resolved_name: str


# ---------------------------
# Comment Create Request
# ---------------------------


class CommentCreateRequest(BaseModel):
    rich_text: list[RichText]
    parent: CommentParent | None = None
    discussion_id: str | None = None
    display_name: CommentDisplayNameInput | None = None
    attachments: list[CommentAttachmentInput] | None = None

    @classmethod
    def for_page(
        cls,
        page_id: str,
        rich_text: list[RichText],
        display_name: CommentDisplayNameInput | None = None,
        attachments: list[CommentAttachmentInput] | None = None,
    ) -> CommentCreateRequest:
        return cls(
            rich_text=rich_text,
            parent=PageCommentParent(page_id=page_id),
            display_name=display_name,
            attachments=attachments,
        )

    @classmethod
    def for_block(
        cls,
        block_id: str,
        rich_text: list[RichText],
        display_name: CommentDisplayNameInput | None = None,
        attachments: list[CommentAttachmentInput] | None = None,
    ) -> CommentCreateRequest:
        return cls(
            rich_text=rich_text,
            parent=BlockCommentParent(block_id=block_id),
            display_name=display_name,
            attachments=attachments,
        )

    @classmethod
    def for_discussion(
        cls,
        discussion_id: str,
        rich_text: list[RichText],
        display_name: CommentDisplayNameInput | None = None,
        attachments: list[CommentAttachmentInput] | None = None,
    ) -> CommentCreateRequest:
        return cls(
            rich_text=rich_text,
            discussion_id=discussion_id,
            display_name=display_name,
            attachments=attachments,
        )


# ---------------------------
# Comment List Request
# ---------------------------


class CommentListRequest(BaseModel):
    block_id: str
    start_cursor: str | None = None
    page_size: int | None = None


# ---------------------------
# User Reference
# ---------------------------


class UserRef(BaseModel):
    object: Literal["user"] = "user"
    id: str


# ---------------------------
# Comment DTO (Response)
# ---------------------------


class CommentDto(BaseModel):
    object: Literal["comment"] = "comment"
    id: str

    parent: CommentParent
    discussion_id: str

    created_time: datetime
    last_edited_time: datetime

    created_by: UserRef

    rich_text: list[RichText] = Field(default_factory=list)
    attachments: list[CommentAttachmentDto] = Field(default_factory=list)
    display_name: CommentDisplayNameDto | None = None


# ---------------------------
# List Response
# ---------------------------


class CommentListResponse(BaseModel):
    object: Literal["list"] = "list"
    results: list[CommentDto] = Field(default_factory=list)
    next_cursor: str | None = None
    has_more: bool = False



from notionary.blocks.schemas import BlockCreatePayload
from notionary.page.content.parser.post_processing.port import PostProcessor


class BlockPostProcessor:
    def __init__(self) -> None:
        self._processors: list[PostProcessor] = []

    def register(self, processor: PostProcessor) -> None:
        self._processors.append(processor)

    def process(
        self, created_blocks: list[BlockCreatePayload]
    ) -> list[BlockCreatePayload]:
        result = created_blocks
        for processor in self._processors:
            result = processor.process(created_blocks)
        return result



from typing import override

from notionary.blocks.rich_text.name_id_resolver.port import NameIdResolver
from notionary.user.client import UserHttpClient
from notionary.user.person import PersonUser


class PersonNameIdResolver(NameIdResolver):
    def __init__(
        self, person_user_factory=None, http_client: UserHttpClient | None = None
    ) -> None:
        if person_user_factory is None:
            person_user_factory = PersonUser
        self.person_user_factory = person_user_factory
        self.http_client = http_client

    @override
    async def resolve_name_to_id(self, name: str | None) -> str | None:
        if not name or not name.strip():
            return None

        name = name.strip()

        try:
            user = await self.person_user_factory.from_name(name, self.http_client)
            return user.id if user else None
        except Exception:
            return None

    @override
    async def resolve_id_to_name(self, user_id: str | None) -> str | None:
        if not user_id or not user_id.strip():
            return None

        try:
            user = await self.person_user_factory.from_id(
                user_id.strip(), self.http_client
            )
            return user.name if user else None
        except Exception:
            return None



from typing import override

from notionary.blocks.schemas import CreateFileBlock, ExternalFileWithCaption
from notionary.page.content.parser.parsers.file_like_block import FileLikeBlockParser
from notionary.page.content.syntax.definition import (
    SyntaxDefinition,
    SyntaxDefinitionRegistry,
)


class FileParser(FileLikeBlockParser[CreateFileBlock]):
    @override
    def _get_syntax(
        self, syntax_registry: SyntaxDefinitionRegistry
    ) -> SyntaxDefinition:
        return syntax_registry.get_file_syntax()

    @override
    def _create_block(self, file_data: ExternalFileWithCaption) -> CreateFileBlock:
        return CreateFileBlock(file=file_data)



from notionary.http.client import NotionHttpClient
from notionary.page.schemas import NotionPageDto


class NotionPageHttpClient(NotionHttpClient):
    def __init__(
        self,
        page_id: str,
    ):
        super().__init__()
        self._page_id = page_id

    async def get_page(self) -> NotionPageDto:
        response = await self.get(f"pages/{self._page_id}")
        return NotionPageDto.model_validate(response)



from typing import override

from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.blocks.schemas import Block, BlockType
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer


class TableRenderer(BlockRenderer):
    MINIMUM_COLUMN_WIDTH = 3

    def __init__(
        self, rich_text_markdown_converter: RichTextToMarkdownConverter | None = None
    ) -> None:
        super().__init__()
        self._rich_text_markdown_converter = (
            rich_text_markdown_converter or RichTextToMarkdownConverter()
        )
        self._table_syntax = self._syntax_registry.get_table_syntax()

    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.TABLE

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        table_markdown = await self._build_table_markdown(context.block)

        if not table_markdown:
            context.markdown_result = ""
            return

        if context.indent_level > 0:
            table_markdown = context.indent_text(table_markdown)

        children_markdown = await context.render_children_with_additional_indent(1)

        if children_markdown:
            context.markdown_result = f"{table_markdown}\n{children_markdown}"
        else:
            context.markdown_result = table_markdown

    async def _build_table_markdown(self, block: Block) -> str:
        if not block.table or not block.has_children or not block.children:
            return ""

        rows = []
        for row_block in block.children:
            if row_block.type != BlockType.TABLE_ROW or not row_block.table_row:
                continue

            row_cells = await self._extract_row_cells(row_block)
            rows.append(row_cells)

        if not rows:
            return ""

        max_columns = max(len(row) for row in rows)
        normalized_rows = self._normalize_row_lengths(rows, max_columns)
        column_widths = self._calculate_column_widths(normalized_rows, max_columns)

        markdown_lines = []

        first_row = normalized_rows[0]
        formatted_first_row = self._format_row(first_row, column_widths)
        markdown_lines.append(formatted_first_row)

        separator_line = self._create_separator_line(column_widths)
        markdown_lines.append(separator_line)

        remaining_rows = normalized_rows[1:]
        for row in remaining_rows:
            formatted_row = self._format_row(row, column_widths)
            markdown_lines.append(formatted_row)

        return "\n".join(markdown_lines)

    def _normalize_row_lengths(
        self, rows: list[list[str]], target_length: int
    ) -> list[list[str]]:
        return [row + [""] * (target_length - len(row)) for row in rows]

    def _calculate_column_widths(
        self, rows: list[list[str]], num_columns: int
    ) -> list[int]:
        widths = [max(len(row[i]) for row in rows) for i in range(num_columns)]
        return [max(width, self.MINIMUM_COLUMN_WIDTH) for width in widths]

    def _format_row(self, cells: list[str], column_widths: list[int]) -> str:
        centered_cells = [cell.center(column_widths[i]) for i, cell in enumerate(cells)]
        delimiter = self._table_syntax.start_delimiter
        return f"{delimiter} {f' {delimiter} '.join(centered_cells)} {delimiter}"

    def _create_separator_line(self, column_widths: list[int]) -> str:
        separators = ["-" * width for width in column_widths]
        delimiter = self._table_syntax.start_delimiter
        return f"{delimiter} {f' {delimiter} '.join(separators)} {delimiter}"

    def _has_column_header(self, block: Block) -> bool:
        if not block.table:
            return False
        return block.table.has_column_header or False

    def _has_row_header(self, block: Block) -> bool:
        if not block.table:
            return False
        return block.table.has_row_header or False

    def _get_table_width(self, block: Block) -> int:
        if not block.table:
            return 0
        return block.table.table_width or 0

    async def _extract_row_cells(self, row_block: Block) -> list[str]:
        if not row_block.table_row or not row_block.table_row.cells:
            return []

        cells = []
        for cell in row_block.table_row.cells:
            cell_text = await self._rich_text_markdown_converter.to_markdown(cell)
            cells.append(cell_text or "")

        return cells



from .data_source import DataSourceNameIdResolver
from .database import DatabaseNameIdResolver
from .page import PageNameIdResolver
from .person import PersonNameIdResolver
from .port import NameIdResolver

__all__ = [
    "DataSourceNameIdResolver",
    "DatabaseNameIdResolver",
    "NameIdResolver",
    "PageNameIdResolver",
    "PersonNameIdResolver",
]



from typing import override

from notionary.blocks.enums import BlockColor
from notionary.blocks.schemas import CreateParagraphBlock, CreateParagraphData
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class SpaceParser(LineParser):
    def __init__(self, syntax_registry: SyntaxDefinitionRegistry) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_space_syntax()

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False

        if self._is_explicit_space_marker(context):
            return True

        return self._is_second_consecutive_empty_line(context)

    def _is_explicit_space_marker(self, context: BlockParsingContext) -> bool:
        return self._syntax.regex_pattern.match(context.line.strip()) is not None

    def _is_second_consecutive_empty_line(self, context: BlockParsingContext) -> bool:
        return context.line.strip() == "" and context.is_previous_line_empty

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        block = self._create_space_block()
        if block:
            context.result_blocks.append(block)

    def _create_space_block(self) -> CreateParagraphBlock:
        paragraph_data = CreateParagraphData(rich_text=[], color=BlockColor.DEFAULT)
        return CreateParagraphBlock(paragraph=paragraph_data)



from enum import StrEnum

from pydantic import BaseModel, Field, model_validator


class UploadMode(StrEnum):
    SINGLE_PART = "single_part"
    MULTI_PART = "multi_part"


class FileUploadStatus(StrEnum):
    PENDING = "pending"
    UPLOADED = "uploaded"
    FAILED = "failed"
    EXPIRED = "expired"


class FileUploadResponse(BaseModel):
    id: str
    created_time: str
    last_edited_time: str
    expiry_time: str | None = None
    upload_url: str | None = None
    archived: bool
    status: FileUploadStatus
    filename: str | None = None
    content_type: str | None = None
    content_length: int | None = None
    request_id: str | None = None


class FileUploadFilter(BaseModel):
    status: FileUploadStatus | None = None
    archived: bool | None = None


class FileUploadListResponse(BaseModel):
    results: list[FileUploadResponse]
    next_cursor: str | None = None
    has_more: bool


class FileUploadCreateRequest(BaseModel):
    filename: str = Field(..., max_length=900)
    content_type: str | None = None
    content_length: int | None = None
    mode: UploadMode = UploadMode.SINGLE_PART
    number_of_parts: int | None = Field(None, ge=1)

    @model_validator(mode="after")
    def validate_multipart_requirements(self):
        if self.mode == UploadMode.MULTI_PART and self.number_of_parts is None:
            raise ValueError("number_of_parts is required when mode is 'multi_part'")
        if self.mode == UploadMode.SINGLE_PART and self.number_of_parts is not None:
            raise ValueError(
                "number_of_parts should not be provided for 'single_part' mode"
            )
        return self

    def model_dump(self, **kwargs):
        data = super().model_dump(**kwargs)
        return {k: v for k, v in data.items() if v is not None}


class FileUploadSendData(BaseModel):
    file: bytes
    part_number: int | None = Field(None, ge=1)


class FileUploadCompleteRequest(BaseModel):
    pass


class FileUploadAttachment(BaseModel):
    file_upload: dict[str, str]
    name: str | None = None

    @classmethod
    def from_id(cls, file_upload_id: str, name: str | None = None):
        return cls(type="file_upload", file_upload={"id": file_upload_id}, name=name)



from pathlib import Path

from pydantic import BaseModel, ConfigDict, Field

from notionary.file_upload.config.constants import (
    NOTION_MULTI_PART_CHUNK_SIZE_MAX,
    NOTION_MULTI_PART_CHUNK_SIZE_MIN,
    NOTION_RECOMMENDED_CHUNK_SIZE,
)


class FileUploadConfig(BaseModel):
    model_config = ConfigDict(frozen=True)

    multi_part_chunk_size: int = Field(
        default=NOTION_RECOMMENDED_CHUNK_SIZE,
        ge=NOTION_MULTI_PART_CHUNK_SIZE_MIN,
        le=NOTION_MULTI_PART_CHUNK_SIZE_MAX,
        description=(
            "The part size (in bytes) for multi-part uploads. Must be within Notion's allowed range (e.g., 5MB-20MB)."
        ),
    )

    max_upload_timeout: int = Field(
        default=300,
        gt=0,
        description="Maximum time in seconds to wait for an upload to complete.",
    )

    poll_interval: int = Field(
        default=2,
        gt=0,
        description="Interval in seconds for polling the upload status.",
    )

    base_upload_path: Path | None = Field(
        default=None,
        description="Optional default base path for resolving relative file uploads.",
    )



from notionary.page.content.parser.post_processing.port import PostProcessor


class MarkdownRenderingPostProcessor:
    def __init__(self) -> None:
        self._processors: list[PostProcessor] = []

    def register(self, processor: PostProcessor) -> None:
        self._processors.append(processor)

    def process(self, markdown_text: str) -> str:
        result = markdown_text
        for processor in self._processors:
            result = processor.process(result)
        return result



from typing import override

from notionary.blocks.schemas import Block, BlockType
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer


class EquationRenderer(BlockRenderer):
    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.EQUATION

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        expression = self._extract_equation_expression(context.block)

        if not expression:
            context.markdown_result = ""
            return

        syntax = self._syntax_registry.get_equation_syntax()
        equation_markdown = (
            f"{syntax.start_delimiter}{expression}{syntax.end_delimiter}"
        )

        if context.indent_level > 0:
            equation_markdown = context.indent_text(equation_markdown)

        children_markdown = await context.render_children_with_additional_indent(1)

        if children_markdown:
            context.markdown_result = f"{equation_markdown}\n{children_markdown}"
        else:
            context.markdown_result = equation_markdown

    def _extract_equation_expression(self, block: Block) -> str:
        if not block.equation:
            return ""
        return block.equation.expression or ""



from typing import Self, cast

from notionary.user.base import BaseUser
from notionary.user.client import UserHttpClient
from notionary.user.schemas import (
    BotUserDto,
    BotUserResponseDto,
    UserResponseDto,
    UserType,
    WorkspaceOwnerType,
)


class BotUser(BaseUser):
    def __init__(
        self,
        id: str,
        workspace_file_upload_limit_in_bytes: int,
        owner_type: WorkspaceOwnerType | None,
        name: str | None = None,
        avatar_url: str | None = None,
        workspace_name: str | None = None,
    ) -> None:
        super().__init__(id=id, name=name, avatar_url=avatar_url)
        self._workspace_name = workspace_name
        self._workspace_file_upload_limit_in_bytes = (
            workspace_file_upload_limit_in_bytes
        )
        self._owner_type = owner_type

    @property
    def workspace_name(self) -> str | None:
        return self._workspace_name

    @property
    def workspace_file_upload_limit_in_bytes(self) -> int:
        return self._workspace_file_upload_limit_in_bytes

    @property
    def owner_type(self) -> WorkspaceOwnerType | None:
        return self._owner_type

    @classmethod
    def _get_expected_user_type(cls) -> UserType:
        return UserType.BOT

    @classmethod
    async def from_current_integration(
        cls,
        http_client: UserHttpClient | None = None,
    ) -> Self:
        client = http_client or UserHttpClient()
        user_dto = await client.get_current_integration_bot()
        return cls.from_dto(user_dto)

    @classmethod
    def from_dto(cls, user_dto: UserResponseDto) -> Self:
        bot_dto = cast(BotUserResponseDto, user_dto)
        bot_data: BotUserDto = bot_dto.bot

        owner_type = bot_data.owner.type if bot_data and bot_data.owner else None
        workspace_name = bot_data.workspace_name if bot_data else None

        limit = 0
        if bot_data and bot_data.workspace_limits:
            limit = bot_data.workspace_limits.max_file_upload_size_in_bytes

        return cls(
            id=bot_dto.id,
            name=bot_dto.name,
            avatar_url=bot_dto.avatar_url,
            workspace_name=workspace_name,
            workspace_file_upload_limit_in_bytes=limit,
            owner_type=owner_type,
        )

    def __repr__(self) -> str:
        return f"BotUser(id={self._id!r}, name={self._name!r}, avatar_url={self._avatar_url!r}, workspace_name={self._workspace_name!r}, workspace_file_upload_limit_in_bytes={self._workspace_file_upload_limit_in_bytes!r}, owner_type={self._owner_type!r})"



from typing import override

from notionary.blocks.rich_text.markdown_rich_text_converter import (
    MarkdownRichTextConverter,
)
from notionary.blocks.schemas import BlockColor, CreateToDoBlock, CreateToDoData
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class TodoParser(LineParser):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry,
        rich_text_converter: MarkdownRichTextConverter,
    ) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_todo_syntax()
        self._syntax_done = syntax_registry.get_todo_done_syntax()
        self._rich_text_converter = rich_text_converter

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False
        return self._is_todo_line(context.line)

    def _is_todo_line(self, line: str) -> bool:
        return (
            self._syntax.regex_pattern.match(line) is not None
            or self._syntax_done.regex_pattern.match(line) is not None
        )

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        block = await self._create_todo_block(context.line)
        if not block:
            return

        await self._process_nested_children(block, context)
        context.result_blocks.append(block)

    async def _process_nested_children(
        self, block: CreateToDoBlock, context: BlockParsingContext
    ) -> None:
        child_lines = self._collect_child_lines(context)
        if not child_lines:
            return

        child_blocks = await self._parse_child_blocks(child_lines, context)
        if child_blocks:
            block.to_do.children = child_blocks

        context.lines_consumed = len(child_lines)

    def _collect_child_lines(self, context: BlockParsingContext) -> list[str]:
        parent_indent_level = context.get_line_indentation_level()
        return context.collect_indented_child_lines(parent_indent_level)

    async def _parse_child_blocks(
        self, child_lines: list[str], context: BlockParsingContext
    ) -> list[CreateToDoBlock]:
        stripped_lines = self._remove_parent_indentation(child_lines, context)
        children_text = self._convert_lines_to_text(stripped_lines)
        return await context.parse_nested_markdown(children_text)

    def _remove_parent_indentation(
        self, lines: list[str], context: BlockParsingContext
    ) -> list[str]:
        return context.strip_indentation_level(lines, levels=1)

    def _convert_lines_to_text(self, lines: list[str]) -> str:
        return "\n".join(lines)

    async def _create_todo_block(self, text: str) -> CreateToDoBlock | None:
        content, checked = self._extract_todo_content(text)
        if content is None:
            return None

        rich_text = await self._convert_to_rich_text(content)
        return self._build_block(rich_text, checked)

    def _extract_todo_content(self, text: str) -> tuple[str | None, bool]:
        done_match = self._syntax_done.regex_pattern.match(text)
        if done_match:
            return done_match.group(1), True

        todo_match = self._syntax.regex_pattern.match(text)
        if todo_match:
            return todo_match.group(1), False

        return None, False

    async def _convert_to_rich_text(self, content: str):
        return await self._rich_text_converter.to_rich_text(content)

    def _build_block(self, rich_text, checked: bool) -> CreateToDoBlock:
        todo_content = CreateToDoData(
            rich_text=rich_text,
            checked=checked,
            color=BlockColor.DEFAULT,
        )
        return CreateToDoBlock(to_do=todo_content)



from .audio import AudioRenderer
from .base import BlockRenderer
from .bookmark import BookmarkRenderer
from .breadcrumb import BreadcrumbRenderer
from .bulleted_list import BulletedListRenderer
from .callout import CalloutRenderer
from .code import CodeRenderer
from .column import ColumnRenderer
from .column_list import ColumnListRenderer
from .divider import DividerRenderer
from .embed import EmbedRenderer
from .equation import EquationRenderer
from .fallback import FallbackRenderer
from .file import FileRenderer
from .heading import HeadingRenderer
from .image import ImageRenderer
from .numbered_list import NumberedListRenderer
from .paragraph import ParagraphRenderer
from .pdf import PdfRenderer
from .quote import QuoteRenderer
from .table import TableRenderer
from .table_of_contents import TableOfContentsRenderer
from .table_row import TableRowHandler
from .todo import TodoRenderer
from .toggle import ToggleRenderer
from .video import VideoRenderer

__all__ = [
    "AudioRenderer",
    "BlockRenderer",
    "BookmarkRenderer",
    "BreadcrumbRenderer",
    "BulletedListRenderer",
    "CalloutRenderer",
    "CodeRenderer",
    "ColumnListRenderer",
    "ColumnRenderer",
    "DividerRenderer",
    "EmbedRenderer",
    "EquationRenderer",
    "FallbackRenderer",
    "FileRenderer",
    "HeadingRenderer",
    "ImageRenderer",
    "NumberedListRenderer",
    "ParagraphRenderer",
    "PdfRenderer",
    "QuoteRenderer",
    "TableOfContentsRenderer",
    "TableRenderer",
    "TableRowHandler",
    "TodoRenderer",
    "ToggleRenderer",
    "VideoRenderer",
]



from typing import override

from notionary.blocks.enums import BlockType
from notionary.blocks.schemas import Block
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer


class ColumnRenderer(BlockRenderer):
    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.COLUMN

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        column_start = self._format_column_start(context.block, context.indent_level)
        children_markdown = await self._render_children_with_indentation(context)

        if children_markdown:
            context.markdown_result = f"{column_start}\n{children_markdown}"
        else:
            context.markdown_result = column_start

    def _format_column_start(self, block: Block, indent_level: int) -> str:
        column_start = self._build_column_start_tag(block)

        if indent_level > 0:
            indent = "    " * indent_level
            column_start = f"{indent}{column_start}"

        return column_start

    def _build_column_start_tag(self, block: Block) -> str:
        delimiter = self._syntax_registry.get_column_syntax().start_delimiter

        if not block.column:
            return delimiter

        width_ratio = block.column.width_ratio
        if width_ratio:
            return f"{delimiter} {width_ratio}"

        return delimiter

    async def _render_children_with_indentation(
        self, context: MarkdownRenderingContext
    ) -> str:
        original_indent = context.indent_level
        context.indent_level += 1

        children_markdown = await context.render_children()

        context.indent_level = original_indent

        return children_markdown
